//
//  designProblem.cpp
//  QPSO

#include "designProblem.hpp"

designProblem::designProblem(){};
designProblem::~designProblem(){};


/*
 
----------------------------------------------------------------------------
                    initializeDesignProbem
----------------------------------------------------------------------------
 
 Arguments:
    nDiscrete         - integer number of discrete factors
    nContinuous       - integer number of continuous factors
    nInteractions     - integer number of interactions
    b                 - Eigen class vector of nominal parameter values
    lbound            - Eigen class vector of lower bounds for the factors
    ubound            - Eigen class vector of upper bounds for the factors
    interactionMatrix - Eigen class matrix of interaction indicators
    optimalityType    - string type of optimality (local vs bayes)
    sensType          - string type of sensitivity function calculation (grid or pso)
    optimalityCrit    - D- or A- optimality. A-optimality still WIP
 
 Description:
    Method takes a set of inputs and initializes a designProblem class object. This
    object can then have various searches run to find an optimal design for the 
    design problem specified.

*/

void designProblem::initializeDesignProbem(int nDiscrete, int nContinuous, int nInteractions, VectorXd b, VectorXd lbound, VectorXd ubound, MatrixXd interactionMatrix, std::string optimalityType, std::string sensType, std::string optimalityCrit){
    
    // Set the basic class attributes from the input
    nDis = nDiscrete; nCont = nContinuous; nInteraction = nInteractions;
    nDesignColumn = nDis + nCont + 1;
    lb.resize( lbound.size() ); lb = lbound;
    ub.resize( ubound.size() ); ub = ubound;
    beta.resize( b.size() ); beta = b;
    optType = optimalityType;
    nParameter = (int)b.size();
    
    // Set the type of search used to verify optimality
    sensitivtyType = sensType;
    
    // Set the optimality criterion to either D or A
    optCrit = optimalityCrit;
    
    // Resize the interaction matrix to be nInteractions x (1 + nDis + nCont)
    interactions.resize(nInteractions, 1 + nDis + nCont);
    interactions = interactionMatrix;
    
}

/*
 
 ----------------------------------------------------------------------------
                            readInDesign
 ----------------------------------------------------------------------------
 
 Arguments:
    filepath       - string of filepath to folder where CSV file is stored.
    filename       - string name of the CSV file

 Description:
    Method to allow the user to read in a saved design in CSV format. The designProblem
    object can then be used as if a search had already been performed. See CSV output of
    designProblem object for appropriate structure of CSV file.
 
 */

void designProblem::readInDesign(std::string filepath, std::string filename){
    
    std::ifstream csv(filepath + filename);
    std::string line;
    std::vector <std::vector<std::string>> items;
    
    // Grab elements from CSV file
    if (csv.is_open()) {
        for (std::string row_line; std::getline(csv, row_line);)
        {
            items.emplace_back();
            std::istringstream row_stream(row_line);
            for(std::string column; std::getline(row_stream, column, ',');)
                items.back().push_back(column);
        }
    }
    else {
        cout << "Unable to open file";
    }
    
    // Copy over the information from the csv into X and p
    nDesignPoint = (int)items.size() - 1;
    nDesignColumn = (int)items[1].size()-1;
    int proportionColumnIndex = (int)items[1].size() - 1;

    // Resize the matrices to the proper size
    X.resize(nDesignPoint, nDesignColumn);
    p.resize(nDesignPoint);
    fullX.resize(nDesignPoint, nParameter);

    for (int iDesignPoint = 1; iDesignPoint <= nDesignPoint; iDesignPoint++){
        
        // Assign the proportion for this design point
        p(iDesignPoint - 1) = std::stod(items[iDesignPoint][proportionColumnIndex]);
        
        // Assign the values of X for this design point
        for (int iCol = 0; iCol < nDesignColumn; iCol++){
            X(iDesignPoint - 1, iCol) = std::stod( items[iDesignPoint][iCol] );
        }
        
    }
    
    // Update the interactions and fullX
    MatrixXd interactionBlock(nDesignPoint, nParameter - nDesignColumn);
    for (int iRow = 0; iRow < nDesignPoint; iRow++){
        for (int iInt = 0; iInt < nInteraction; iInt++){
            double interactionValue = 1.0;
            for (int iCol = 1; iCol < nDesignColumn; iCol++){
                if (interactions(iInt, iCol) == 1){
                    interactionValue *= interactions(iInt, iCol) * X(iRow, iCol);
                }
            }
            interactionBlock(iRow, iInt) = interactionValue;
        }
    }
    fullX << X, interactionBlock;

}

/*
 
 ----------------------------------------------------------------------------
                            updateArchive
 ----------------------------------------------------------------------------
 
 Description:
    Method to take the current design matrix and proportion allocation and
    save it to the archive.
 
 */

void designProblem::updateArchive(){
    
    xArchive.resize( X.row(1).size(), X.col(1).size() );
    pArchive.resize( p.size() );
    
    xArchive = X;
    pArchive = p;
    
}

/*
 
 ----------------------------------------------------------------------------
                            qpso
 ----------------------------------------------------------------------------
 
 Arguments:
    num_swarm                   - integer number of quantum-behaved particle swarms to use
    num_parts                   - integer number of particles in each swarm
    maxit                       - integer maximum number of iterations
    maxresets                   - integer maximum number of times the swarm will reset if it
                                    fails to find a design of the desired quality
    lower_efficiency_bound      - lower bound on the relative efficiency of the obtained design
                                    to the theoretically D-optimal design required to terminate.
                                    Not used for bayesian approach.
    link                        - link function to relate linear predictor to outcome
    rowLimit                    - maximum number of support points. For exact designs this is N.
    priors                      - string vector of priors. Currently handles uniform "U(x,y)" and
                                    Normal "N(x,y)" distributions.
    breedIterNum                - integer number of iterations between elitist breeding mutations
    outpath                     - filepath to the folder where the design is to be stored
    prefix                      - name for the csv file with the design
    logging                     - boolean, if 1, then pbests for each iteration are saved. Do not use
                                    if using lots of particles or swarms (will create huge file)
    allocation                  - string: "exact" for exact designs, "approx" for approximate designs
    printIterBest               - if TRUE, each swarms gbest is printed at the end of each iteration
    printFinalInfo              - if TRUE, the final design is printed to the console
    saveFinalDesign             - if TRUE, the final design is saved to outpath/prefix_gbest_design.csv
 
 Description: 
    This is the main algorithm function. It runs the d-QPSO search under the specified settings
    by creating a qhabitat. It then performs a seach of the sensitivity function (for local optimality)
 
 */

void designProblem::qpso(int num_swarm, int num_parts, int maxit, int maxresets, double lower_efficiency_bound, std::string link, int rowLimit, std::vector<string> priors, int breedIterNum, string outpath, string prefix, bool logging, std::string allocation, bool printIterBest, bool printFinalInfo, bool saveFinalDesign){
    
    // Create the integration approximation for pseudo-Bayesian approach
    integrationAbscissas *IA = new integrationAbscissas;
    if (optType == "bayes"){
        IA->performPriorConversion(priors);
        IA->generateRadialQuantities(nParameter);
        IA->generateSimplexMethod(nParameter);
        IA->generateOrthoRotations(nParameter);
        IA->generateThetaSearchPoints(nParameter);
        IA->performPriorApproximation();
    }
    
    // Set up parameter information for EW designs
    VectorXd lowerLimit( nParameter ); VectorXd upperLimit( nParameter );
    
    // Set up the search habitat and run the search
    qhabitat searchHabitat;
    searchHabitat.initializeHabitat(num_swarm, num_parts, maxit, IA, lb, ub, beta, nDis, nCont, nInteraction,  interactions, rowLimit, breedIterNum, 1, optType, lower_efficiency_bound, lowerLimit, upperLimit, logging, allocation, sensitivtyType, printIterBest, optCrit);
    searchHabitat.runEntireSearch();
    
    // Find the maximum of the sensitivity function using a QPSO algorithm
    if (optType == "local" && sensitivtyType == "pso"){
        equivswarm checker;
        int mm = 1000;
        checker.intitilizeEquivSwarm(searchHabitat.getGbestFisherInformation(), beta, nCont, nDis, lb, ub, 75, mm, nInteraction, interactions);
        bool conv = false;
        for (int iter = 0; iter < mm; iter++){
            checker.updateMBest();
            checker.updateIteratorLocal();
            conv = checker.checkConvergence();
            if (conv){iter = mm;}
        }
    }

    // Resize the design matrix to the appropriate size and store the best result from the search
    X.resize(rowLimit, nDesignColumn); p.resize(rowLimit);
    X = searchHabitat.getHBestX();
    p = searchHabitat.getHBestP();
    fullX.resize(rowLimit, nDesignColumn + nInteraction);
    
    // Update the interactions and fullX
    MatrixXd interactionBlock(rowLimit, nParameter - nDesignColumn);
    for (int iRow = 0; iRow < rowLimit; iRow++){
        for (int iInt = 0; iInt < nInteraction; iInt++){
            double interactionValue = 1.0;
            for (int iCol = 1; iCol < nDesignColumn; iCol++){
                //cout << interactionMatrix(iInt, iCol) * X(iRow, iCol) << "\n";
                if (interactions(iInt, iCol) == 1){
                    interactionValue *= interactions(iInt, iCol) * X(iRow, iCol);
                }
            }
            interactionBlock(iRow, iInt) = interactionValue;
        }
    }
    fullX << X, interactionBlock;
    
    // Write the final design to a CSV file
    if (saveFinalDesign){
        searchHabitat.saveHBestDesign(outpath, prefix);
    }
    
    // Get stats for obtained design
    nDesignPoint = (int)X.col(1).size();
    fitnessArchive = searchHabitat.getHBest();
    lowestBoundAchieved = searchHabitat.getLEB();
    
    // If user requested, print out the final information about the design
    if (printFinalInfo){
        searchHabitat.printHBestDesign();
        cout << "The fitness was: " << searchHabitat.getHBest();
        cout << "The lower bound on the efficiency: " << getLEBAchieved() << endl;
    }
    
}

/*
 
 ----------------------------------------------------------------------------
                            setFullDesign
 ----------------------------------------------------------------------------
 
 Arguments:
    Xnew          - Eigen matrix, design matrix
    pnew          - Eigen vector object of the proportions for each support point
 
 Description: 
    Sets the stored X and p to Xnew and pnew.
 
 */

void designProblem::setFullDesign(MatrixXd Xnew, VectorXd pnew){
    nDesignColumn = 1 + nDis + nCont;
    nDesignPoint = (int)Xnew.col(1).size();
    
    // Resize the matrices to the proper size
    X.resize(nDesignPoint, nDesignColumn);
    p.resize(nDesignPoint);
    fullX.resize(nDesignPoint, nParameter);
    
    X = Xnew;
    p = pnew;
    fullX = Xnew;
};

/*
 
 ----------------------------------------------------------------------------
                        Printing Methods
 ----------------------------------------------------------------------------
 
Method List:
 printX                     - prints the design matrix for the d-QPSO algorithm-generated design
 printP                     - prints the proportions allocated to each support point
 printDesign                - prints both X and p (for easy design viewing)
 printLowestBoundAchieved   - prints the best bound achieved
 printArchivedDesign        - prints the archived design (otherwise same as printDesign method)
 
 
*/

void designProblem::printX(){
    for (int iRow = 0; iRow < nDesignPoint; iRow++){
        for (int iCol = 0; iCol < nDesignColumn; iCol++){
            cout << X(iRow, iCol) << " ";
        }
    }
}

void designProblem::printP(){
    for (int iRow = 0; iRow < nDesignPoint; iRow++){
        cout << p(iRow) << endl;
    }
};

void designProblem::printDesign(){
    for (int iRow = 0; iRow < nDesignPoint; iRow++){
        for (int iCol = 0; iCol < nDesignColumn; iCol++){
            cout << X(iRow, iCol) << " ";
        }
        cout << p(iRow) << endl;
    }
};

void designProblem::printLowestBoundAchieved(){
    cout << "The lowest bound achieved is: " << lowestBoundAchieved << endl;
}

void designProblem::printArchivedDesign(){
    cout << "The Archived Design is: " << endl;
    for (int iRow = 0; iRow < xArchive.col(1).size(); iRow++){
        for (int iCol = 0; iCol < nDesignColumn; iCol++){
            cout << xArchive(iRow, iCol) << " ";
        }
        cout << pArchive(iRow) << endl;
    }
    cout << "With a corresponding bound of " << lowestBoundAchieved << endl;
}

/*
 
 ----------------------------------------------------------------------------
                        incrementSetting
 ----------------------------------------------------------------------------
 
 Arguments:
    setting          - current design point setting
    nSearchValue     - vector of number of different values for each factor (depends on grid size)
    iFactor          - index of what factor is currently being incremented
 
 Description: Recursive helper function for the grid search of the sensitivity function. Used
    to select the next candidate design point.
 
 */


std::vector<int> incrementSetting(std::vector<int> setting, std::vector<int> nSearchValue, int iFactor){
    
    // Handle case where we are at the end
    if ( setting[iFactor] == nSearchValue[iFactor]-1 ){
        if (iFactor > 0){
            return incrementSetting( setting, nSearchValue, iFactor - 1 );
        }
        else{
            setting[0] = -999;
            return(setting);
        }
    }
    // Increment
    else{
        std::vector<int> newSetting; newSetting.resize( setting.size() ); newSetting = setting;
        newSetting[iFactor] = setting[iFactor] + 1;
        // reset everything to the right
        if ( iFactor < nSearchValue.size()-1 ){
            for (int iAbove = iFactor+1; iAbove < nSearchValue.size(); iAbove++ ){
                newSetting[iAbove] = 0;
            }
        }
        return(newSetting);
    }
    
}

/*
 
 ----------------------------------------------------------------------------
                        fullSensitivityCheck
 ----------------------------------------------------------------------------
 
 Arguments:
    resolution     - resolutions of the grid used to break up continuous factors. For example,
                        resolution = 100 will consider 100 equally spacedsettings for each continuous
                        factor along the factor range. Note that as the number of continuous factors
                        increases this can become very time consuming.
 
 Description: 
    Function to evaluate the sensitivity function of the obtained design. This function finds the
    maximum of the sensitivity function of a design using a grid search and then outputs the efficiency 
    lower bound.
 
 Output:
    elb             - double, lower bound on the relative D-efficiency of the obtained design to the
                        theoretical locally D-optimal design.
 
 */

double designProblem::fullSensitivityCheck(int resolution){
    
    // Step 1: obtain inverse of the fisher information matrix
    
    // Allocate space for the fisher information matrix, its inverse, W, and linear predictor
    MatrixXd fisherInfo(nParameter, nParameter);
    MatrixXd invFisherInfo(nParameter, nParameter);
    MatrixXd W(nDesignPoint, nDesignPoint);
    VectorXd xb(nDesignPoint);
    
    // Calculate W Matrix
    for (int iRow = 0; iRow < nDesignPoint; iRow++){
        for (int iCol = 0; iCol < nDesignPoint; iCol++){
            W(iRow, iCol) = 0;
        }
    }
    xb = fullX * beta;
    for (int ii = 0; ii < nDesignPoint; ii++){
        W(ii,ii) = exp(xb(ii)) / ( ( 1 + exp(xb(ii)) ) * ( 1+exp(xb(ii)) ) );
        W(ii, ii) = W(ii, ii) * p(ii);
    }
    
    // Obtain the Fisher Information Matrix and its inverse
    fisherInfo = fullX.transpose() * W * fullX;
    invFisherInfo = fisherInfo.inverse();
    
    // Step 2: Set up the grid of points to search over
    
    // Store the number of factors and pre-allocate space for the search points
    int nFactor = (int)lb.size();
    double leb = 999;
    std::vector<int> nSearchValue(nFactor);
    
    // Fill out the number of search values. Will be 2 for discrete factors and RESOLUTION for continuous
    for (int iFactor = 0; iFactor < nFactor; iFactor++){
        if (iFactor < nDis){ nSearchValue[iFactor] = 2; }
        else {  nSearchValue[iFactor] = resolution; }
    }
    
    // Now create an array of the size of the search values, so 2 x 2 x ... x 2 x res x res x ... x res
    std::vector< std::vector<double> > candValues;
    candValues.resize(nFactor);
    // Loop through the array and populate it with the values
    int nPoints = 0;
    double incr = 0;
    for (int iFactor  = 0; iFactor < nFactor; iFactor++){
        nPoints = nSearchValue[iFactor];
        // Resize this part of the array to the correct size (nPoints)
        candValues[iFactor].resize(nPoints);
        // Populate for a discrete factor
        if (nPoints == 2){ candValues[iFactor][0] = -1.0; candValues[iFactor][1] = 1.0; }
        // Populate for a continuous factor
        else{
            // get increment size
            incr = ( ub(iFactor) - lb(iFactor) ) / (resolution-1);
            // loop through and add values to search over
            double lowestValue = lb[iFactor];
            for (int iGrid = 0; iGrid < resolution; iGrid++){
                candValues[iFactor][iGrid] = lowestValue + ( iGrid * incr + 0.0 );
            }
        }
    }
    
    // Step 3: Loop over candidate points, find maximum value of sensitivity function
    
    // Now need to evaluate each of the prod( nSearchValue ) possible design points
    // initialize a vector to track current position
    std::vector<int> currentSetting; currentSetting.resize(nFactor);
    for (int ii = 0; ii < nFactor; ii++){ currentSetting[ii] = 0; }
    // Create vector to store current candidate design point and assign the intercept
    VectorXd xPoint (nParameter); xPoint(0) = 1;
    // Create storage variables
    double wPoint = 0;
    double xbPoint = 0;
    double sensValue = 0;
    double maxSens = -9999;
    // Obtain the current setting
    currentSetting = incrementSetting(currentSetting, nSearchValue, nFactor-1);
    
    while (currentSetting[0] > -998){
        
        // set the design point to the current setting
        for (int iFactor = 1; iFactor < nDesignColumn; iFactor++){
            xPoint(iFactor) = candValues[iFactor-1][currentSetting[iFactor-1]];
        }
        
        // Loop over the interactions
        for (int iInteract = 0; iInteract < nInteraction; iInteract++){
            double interactionValue = 1.0;
            for (int iCol = 1; iCol < nDesignColumn; iCol++){
                //cout << interactionMatrix(iInt, iCol) * X(iRow, iCol) << "\n";
                if (interactions(iInteract, iCol) == 1){
                    interactionValue *= interactions(iInteract, iCol) * xPoint(iCol);
                }
            }
            xPoint(iInteract + nDesignColumn) = interactionValue;
        }
        
        // Calculate the value of w at this design point
        xbPoint = xPoint.transpose() * beta;
        wPoint = exp(xbPoint) / ( ( 1 + exp(xbPoint) ) * ( 1+exp(xbPoint) ) );
        
        // Evaluate the sensitivity function
        sensValue = wPoint * xPoint.transpose() * invFisherInfo * xPoint - nParameter;
        
        if (sensValue > maxSens){maxSens = sensValue;}
        
        // Obtain the current setting for the next point
        currentSetting = incrementSetting(currentSetting, nSearchValue, nFactor-1);
        
    }
    
    // Step 4: Use the maximum value of the sensitivity function to obtain the lower bound
    
    leb = exp(-maxSens / nParameter);
    
    if (leb > lowestBoundAchieved){
        lowestBoundAchieved = leb;
        updateArchive();
    }
    
    cout << "The worst was: " << maxSens << endl;
    cout << "The corresponding lbe is: " << leb << endl;
    
    return(leb);
}

