//
//  designProblem.hpp
//  QPSO


#ifndef designProblem_hpp
#define designProblem_hpp

#include "qpso.h"

/*
 designProblem
 
 -------------------------------------------------------------------------------------------
 Class keeps track of the overall type of design problem the user has set up
 -------------------------------------------------------------------------------------------
 
 Class Attributes:
    X                    :    The obtained experimental design matrix without the interactions calculated
    fullX                :    The obtained experimental design matrix WITH interactions included
    interactions         :    List of the interactions
    xArchive             :    Saved design from a previous run (not currently in use)
    pArchive             :    Saved allocated from a previous run (not currently in use)
    beta                 :    Vector of nominal parameter values
    lb                   :    Vector of lower bounds for each factor
    ub                   :    Vector of upper bounds for each factor
    p                    :    Vector of allocations to support points.
    fitnessArchive       :    Saved best fitness
    lowestBoundAchieved  :    Best bound found using sensitivity function
    nDis                 :    Number of discrete factors in design problem
    nCont                :    Number of continuous factors in design problem
    nInteraction         :    Number of interactions in the design problem
    nParameter           :    Number of parameters in the model
    nDesignPoint         :    Number of support points
    nDesignColumn        :    The number of columns in the design matrix
    optType              :    Type of optimality requested (local of bayes)
    optCrit              :    D- or A- Optimality, currently A is incomplete
    sensitivityType      :    Search method used to find the maximum of the sensitivity function (grid, pso)
 */

class designProblem{
    
public:
    
    int nDis, nCont, nInteraction, nParameter, nDesignPoint, nDesignColumn;
    VectorXd beta, lb, ub, p, pArchive;
    MatrixXd X, interactions, fullX, xArchive;
    std::string optType, optCrit, sensitivtyType;
    double lowestBoundAchieved = 0.00;
    double fitnessArchive;
    
    // Class Constructor and initializer
    designProblem();
    void initializeDesignProbem(int nDiscrete, int nContinuous, int nInteractions, VectorXd b, VectorXd lbound, VectorXd ubound, MatrixXd interactionMatrix, std::string optimalityType, std::string sensitivtyType, std::string optimalityCrit);
    
    // Search Functions
    void qpso(int num_swarm, int num_parts, int maxit, int maxresets, double lower_efficiency_bound, std::string link, int rowLimit, std::vector<string> priors, int breedIterNum, string outpath, string prefix, bool logging, std::string allocation, bool printIterBest, bool printFinalInfo_set, bool saveFinalDesign);
    
    // Getter functions
    int getNDiscrete(){ return nDis; }
    int getNContinuous(){ return nCont; }
    VectorXd getBeta(){ return beta; }
    double getLEBAchieved(){return lowestBoundAchieved;}
    MatrixXd getDesign(){ return X; }
    VectorXd getP(){return p;}
    
    // Setter Function
    void readInDesign(std::string filepath, std::string filename);
    void updateArchive();
    void setFullDesign(MatrixXd X, VectorXd p);
    
    // Printing Methods
    void printX();
    void printP();
    void printDesign();
    void printLowestBoundAchieved();
    void printArchivedDesign();
    
    // Sensitivity Function calculations
    double fullSensitivityCheck(int resolution);
    double getHBest(){return( fitnessArchive );}
    
    ~designProblem();
    
};


#endif /* designProblem_hpp */
