//
//  equivparticle.cpp
//  QPSO

#include "equivparticle.hpp"

equivparticle::equivparticle(){};
equivparticle::~equivparticle(){};

/*
 
 -------------------------------------------------------------------------------------------
                            intitilizeEquivParticle
 -------------------------------------------------------------------------------------------
 
 Arguments:
    invFisherInformation   :     Eigen matrix object containing the inverse of the Fisher Information matrix
                                    for the design whose sensitvity function is to be checked
    b                      :     Eigen vector object containing the nominal parameter values
    nContinuous            :     integer number of continuous factors
    nDiscrete              :     integer number of discrete factors
    lbd                    :     Eigen vector object of lower bounds for each factor
    ubd                    :     Eigen vector object of upper bounds for each factor
    nparam                 :     integer number of parameters in the model
    nInteractions          :     integer number of interactions in the model
    interact               :     Eigen matrix object of interactions
 
 Description:
    Method to initialize a particle class object searching for the maximum of the sensitivity function to a starting position. Each particle is a single candidate design point.

*/

void equivparticle::intitilizeEquivParticle(MatrixXd invFisherInformation, VectorXd b, int nContinuous, int nDiscrete, VectorXd lbd, VectorXd ubd, int nparam, int nInteractions, MatrixXd interact){
    
    nParameter = nparam;
    
    // Set the inverse Fisher information
    invFisherInfo.resize( nParameter, nParameter );
    invFisherInfo = invFisherInformation;
    
    // Set the lower and upper bounds
    ub.resize(ubd.size()); lb.resize(lbd.size());
    lb = lbd;
    ub = ubd;
    
    // Set the vector of beta coefficients
    beta = b;
    
    // Set the number of discrete and continuous factors
    nCont = nContinuous; nDis = nDiscrete;
    nCol = nContinuous + nDiscrete + 1;
    
    // Interaction information
    nInt = nInteractions;
    interactions.resize(nInt, nCol);
    interactions = interact;
    fullX.resize(nCol+nInt);
    
    // Resize the candidate design point and the pbest candidate design point
    X.resize( nCol );
    pbestX.resize( nCol );
    localAtt.resize( nCol );
    localAtt(0) = 1;
    
    
    //----------------------------------------------\\
    //          Generate Initial Values             \\
    
    std::random_device rdm;
    std::mt19937 generator(rdm());
    std::uniform_real_distribution<double> pdist(0,1);
    
    // fill out the intercept
    X(0) = 1;
    
    // fill out the discrete factors
    for (int iFact = 1; iFact < 1+nDis; iFact++){
        double randVal = pdist(generator);
        if ( randVal > 0.5 ){ X(iFact) = 1;}
        else { X(iFact) = -1;}
        pbestX(iFact) = X(iFact);
    }
    
    // fill out the continuous factors
    for (int iCol = nDis+1; iCol < nCol; iCol++){
        std::uniform_real_distribution<double> distribution( lb(iCol-1), ub(iCol-1) );
        X(iCol) = distribution(generator);
        pbestX(iCol) = X(iCol);
    }
    
    // Update the full design for this design point
    updateInteractions();
    
    // Calculate w for this design point
    updateW();
    
}

/**************************************************************************
 **************************************************************************
                        Update Methods
 **************************************************************************
 **************************************************************************/


/*
 
 -------------------------------------------------------------------------------------------
                                updateInteractions
 -------------------------------------------------------------------------------------------
 
 Description:
    Method to take the design matrix class attribute, X, and calculate the full design including interactions, fullX. Both X and fullX are Eigen matrix class objects.
 
 */


void equivparticle::updateInteractions(){
    
    // Calculate the interactions
    VectorXd interactionBlock(nInt);
    for (int iInt = 0; iInt < nInt; iInt++){
        double interactionValue = 1.0;
        for (int iCol = 1; iCol < nCol; iCol++){
            if (interactions(iInt, iCol) == 1){
                interactionValue *= interactions(iInt, iCol) * X(iCol);
            }
        }
        interactionBlock(iInt) = interactionValue;
    }
    
    // Update fullX based on whether or not there are any interactions
    if (nInt > 0){
        fullX << X, interactionBlock;
    }
    else{
        fullX << X;
    }

}

/*
 
 -------------------------------------------------------------------------------------------
                                    updateW
 -------------------------------------------------------------------------------------------
 
 Description:
    Method to update W based on the linear predictor X and the link function, note that only logit has
    been implemented currently.
 
 */

void equivparticle::updateW(){
    
    // Case 1: Logit link function
    std::string link = "logit";
    if (link == "logit"){
        double xb = fullX.transpose() * beta;
        w = exp(xb) / ( ( 1 + exp(xb) ) * ( 1+exp(xb) ) );
    }
    
}


/*
 
 -------------------------------------------------------------------------------------------
                                  calcFitness
 -------------------------------------------------------------------------------------------
 
 Description:
    Method to calculate the fitness of the particle. For this search, the fitness is the value of the sensitivity function evaluated at the current design point.
 
*/


void equivparticle::calcFitness(){
    fitness = w * fullX.transpose() * invFisherInfo * fullX - nParameter;
}


/*
 
 -------------------------------------------------------------------------------------------
                                checkImprovement
 -------------------------------------------------------------------------------------------
 
 Description:
    Method to check if the current fitness is better than the old stored fitness. That is, this method checks to see if the current design point results in a larger value for the sensitivity function that the current pbest. If so, the pbest is updated.
 
*/

void equivparticle::checkImprovement(){
    
    if (fitness > pbest){
        pbest = fitness;
        pbestX = X;
    }
    
}

/*
 
 -------------------------------------------------------------------------------------------
                                calcAttractor
 -------------------------------------------------------------------------------------------
 
 Arguments:
    gbestX   :     Eigen vector object of the best candidate design point found by the entire swarm
 
 Description:
    Method to calculate the local attractor for the particle. The local attractor is based on the distance the particle currently is from the gbest and pbest positions.
 
 */

void equivparticle::calcAttractor(VectorXd gbestX){
    
    std::random_device rdm;
    std::mt19937 generator(rdm());
    std::uniform_real_distribution<double> pdist(0,1);
    
    // Local attractor for the discrete factors
    for (int iCol = 1; iCol < nDis+1; iCol++){
        // Using uniform crossover
        double rid = pdist(generator);
        if (rid <= 0.5){ localAtt(iCol) = pbestX(iCol); }
        else { localAtt(iCol) = gbestX(iCol); }
    }
        
    // Local attractor for the continuous factors
    for (int iCol = 1+nDis; iCol < nCol; iCol++){
        double rid = pdist(generator);
        double Rid = pdist(generator);
        double phi = ( 2.0*rid ) / (2.0*rid + 2.0*Rid);
        localAtt(iCol) = phi * pbestX(iCol) + (1.0 - phi)*(gbestX(iCol));
    }
    
}

/*
 
 -------------------------------------------------------------------------------------------
                                    updatePosition
 -------------------------------------------------------------------------------------------
 
 Arguments:
    mbestX          - Eigen vector object storting the mean best design point
    alpha           - double value of the expansion contraction coefficent
 
 Description:
    Method to update the position of the particle based on the mean best position
    and the expansion contraction coefficient
 
 */

void equivparticle::updatePosition(VectorXd mbestX, double alpha){
    
    std::random_device rdm;
    std::mt19937 generator(rdm());
    std::uniform_real_distribution<double> pdist(0,1);
    
    double sign = -1.0;
    
    double F = 0;
    int Rd = 0;
    int xorprod = 0;
    int binarySpaceAdd = 0;
    
    double uRand = 0; double randV = 0; double addToLocalAtt;
    
    // Update for the discrete factors
    for (int iCol = 1; iCol < nDis+1; iCol++){

        // Performing the update using bit-wise exclusive operator
        randV = 1 / pdist(generator);
        if (randV < 1){F = randV;} else {F = 1;}
        uRand = pdist(generator);
        if (uRand < F){Rd = 1;} else{Rd = 0;}
            
        if (mbestX(iCol) == X(iCol)){xorprod = 0;} else{xorprod = 1;}
        if (xorprod == Rd){binarySpaceAdd = 1;} else{binarySpaceAdd = -1;}
            
        if (localAtt(iCol) == binarySpaceAdd){ X(iCol)  = -1;}
        else { X(iCol) = 1;}
            
    }
        
    // Update for the continuous factors
    for (int iCol = 1+nDis; iCol < nCol; iCol++){
        // Perform the draw
        randV = pdist(generator);
        if (randV >= 0.5){sign = -1.0;} else{sign = 1.0;}
        
        uRand = pdist(generator);
        addToLocalAtt = sign * alpha * abs( X(iCol) -  mbestX(iCol)) * log(1.0 / uRand);
        X(iCol) = localAtt(iCol) + addToLocalAtt;
    }
    
}

/*
 
 -------------------------------------------------------------------------------------------
                                checkPositionBounds
 -------------------------------------------------------------------------------------------
 
 Description:
    Method to make sure the position of each factor in the design point is within
    the factor bounds.
 
 */

void equivparticle::checkPositionBounds(){
    
    // fix for the continuous factors
    for (int iCol = 1+nDis; iCol < nCol; iCol++){
        if ( X(iCol) < lb(iCol-1) ){ X(iCol) = lb(iCol-1) + 0.0; }
        if ( X(iCol) > ub(iCol-1) ){ X(iCol) = ub(iCol-1) + 0.0;  }
    }
    
    updateInteractions();
    
}








