//
//  equivparticle.hpp
//  QPSO

#ifndef equivparticle_hpp
#define equivparticle_hpp

#include "qpso.h"

/*
 EQUIVPARTICLE
 
 -------------------------------------------------------------------------------------------
 Each equiv particle searches for a solution that maximizes the value of the sensitivity function
 -------------------------------------------------------------------------------------------
 
 Class Attributes:
 invFisherInfo  :    The inverse fisher information. Used in calculation of sensitivity function
 interactions   :    Matrix where each row is an interaction
 beta           :    Vector of nominal parameter values
 lb             :    The lower limit for each design point. Note that this does not include the intercept.
 ub             :    The upper limit for each design point. Note that this does not include the intercept.
 X              :    Vector containing the current design point for the particle position
 fullX          :    Vector containing the current design point WITH interaction
 pbestX         :    Vector of the design point found by this particle with the largest returned sensitivity fn
 localAtt       :    Vector containing the point that this particle is attracted to; weighted avg of gbest, mbest
 nCont          :    int number of continuous factors
 nDis           :    int number of discrete factors
 nInt           :    int number of interactions
 nCol           :    int number of columns of the design matrix XXX does not currently appear to be used
 nParameter     :    int number of parameters in the model, should be the same as the length of beta
 fitness        :    double is sensitvitty value at current design point
 w              :    value of w for the current design point
 pbest          :    double keeps track of the largest value of the sensitivity function
 
 */

class equivparticle{
    
public:
    
    // Class attributes
    MatrixXd invFisherInfo; MatrixXd interactions;
    VectorXd beta, lb, ub, X, fullX, pbestX, localAtt;
    double pbest = -9999999.9; double w = 0; double fitness = 0;
    int nCont, nDis, nInt, nParameter, nCol;
    
    // Class Constructor and initializer
    equivparticle();
    void intitilizeEquivParticle(MatrixXd invFisherInformation, VectorXd b, int nContinuous, int nDiscrete, VectorXd lbd, VectorXd ubd, int nparam, int nInteraction, MatrixXd interact);
    
    // Updating Methods
    void updateW();
    void calcFitness();
    void checkImprovement();
    void calcAttractor(VectorXd gbestX);
    void updatePosition(VectorXd mbestX, double alpha);
    void updateInteractions();
    void checkPositionBounds();
    
    // Getter Methods
    double getPbest(){return pbest;}
    double getFitness(){return fitness;}
    VectorXd getX(){return X;}
    VectorXd getPbestX(){return pbestX;}
    
    // Printing Methods
    
    ~equivparticle();
    
};


#endif /* equivparticle_hpp */
