//
//  equivswarm.cpp
//  QPSO


#include "equivswarm.hpp"

equivswarm::equivswarm(){};
equivswarm::~equivswarm(){};

/*
 
 -------------------------------------------------------------------------------------------
                               intitilizeEquivSwarm
 -------------------------------------------------------------------------------------------
 
 Arguments:
    fisherInformation   - Eigen matrix containing the fisher information of the design being tested
    b                   - Eigen vector of the nominal values
    nContinuous         - integer number of continuous factors
    nDiscrete           - integer number of discrete factors
    lbd                 - Eigen vector of the lower bounds for each factor
    ubd                 - Eigen vector of the upper bounds for each factor
    nParticle           - integer number of particles for the equivalence check swarm
    maxiter             - integer maximum number of iterations
    nInteractions       - integer number of interactions in the model
    interactionMat      - Eigen matrix where each row specifies an interaction
 
 Description:
    Method to initialize the equivalence swarm. Generates a swarm with nParticle particles,
    where each is a candidate design point to maximize the sensitivity function.
 
 */

void equivswarm::intitilizeEquivSwarm(MatrixXd fisherInformation, VectorXd b, int nContinuous, int nDiscrete, VectorXd lbd, VectorXd ubd, int nParticle, int maxiter, int nInteractions, MatrixXd interactionMat){
    
    // Set the lower and upper bounds
    ub.resize(ubd.size()); lb.resize(lbd.size());
    lb = lbd;
    ub = ubd;
    
    // Set the vector of beta coefficients
    nParameter = (int)b.size();
    beta.resize( nParameter );
    beta = b;
    gbestX.resize( lbd.size() );
    mbestX.resize( lbd.size() );
    
    // Set up the interactions
    nInt = nInteractions;
    interactions.resize(nInt, 1 + nContinuous + nDiscrete);
    interactions = interactionMat;
    
    // Set the inverse Fisher information
    invFisherInfo.resize( nParameter, nParameter );
    invFisherInfo = fisherInformation.inverse();
    
    // Set the number of discrete and continuous factors
    nCont = nContinuous; nDis = nDiscrete;
    
    // Set the number of particles
    nPart = nParticle;
    fracPart = 1.0 / nPart;
    
    // Set the maximum number of iterations
    maxit = maxiter;
    
    // Resize the swarm size to the number of particles
    eqSwarm = new equivparticle *[nPart];
    
    // Initialize each of the particles
    for (int part = 0; part < nPart; part++){
        eqSwarm[part] = new equivparticle();
        eqSwarm[part]->intitilizeEquivParticle(invFisherInfo, beta, nCont, nDis, lb, ub, nParameter, nInt, interactions);
        eqSwarm[part]->calcFitness();
        eqSwarm[part]->checkImprovement();
        if (eqSwarm[part]->getPbest() > gbest){
            gbest = eqSwarm[part]->getFitness();
            gbestIndex = part;
        }
    }
    gbestX = eqSwarm[gbestIndex]->getX();
    
}

/*
 
 -------------------------------------------------------------------------------------------
                                updateMBest
 -------------------------------------------------------------------------------------------
 
 Description:
    Method to update the mean best position by taking the average of the pbest position
    for each particle in the swarm.
 
 */

void equivswarm::updateMBest(){
    
    iter = iter + 1;
    
    // Calculate mbest from the previous iteration
    mbestX = fracPart * eqSwarm[0]->getPbestX();
    for (int particle = 1; particle < nPart; particle++){
        mbestX = mbestX + fracPart * eqSwarm[particle]->getPbestX();
    }
    
    // if mbest at discrete < 0, then mbest = -1, else mbest = 1
    for (int iCol = 1; iCol < 1+nDis; iCol++){
        if (mbestX(iCol) < 0){ mbestX(iCol) = -1; }
        else { mbestX(iCol) = 1; }
    }
    
}

/*
 
 -------------------------------------------------------------------------------------------
                                updateIteratorLocal
 -------------------------------------------------------------------------------------------
 
 Description:
    Method to update a single iteration in the search for the maximum of the sensitivity fn.
 
 */

void equivswarm::updateIteratorLocal(){
    
    minval = 999;
    alpha = 1.2 - 0.7 * ( (iter - 1) / (maxit - 1) );
    
    for (int i=0;i<nPart;i++){
        
        eqSwarm[i]->calcAttractor(gbestX);
        eqSwarm[i]->updatePosition(mbestX, alpha);
        eqSwarm[i]->checkPositionBounds();
        eqSwarm[i]->updateW();
        eqSwarm[i]->calcFitness();
        eqSwarm[i]->checkImprovement();
        
        if (eqSwarm[i]->getFitness() > gbest){ gbestIndex = i; updateGbest = true; gbest = eqSwarm[gbestIndex]->getFitness();}
        if (eqSwarm[i]->getPbest() < minval){ minval = eqSwarm[i]->getPbest(); worstIndex = i;}
        
        
    }
    
    if (updateGbest){
        gbestX = eqSwarm[gbestIndex]->getPbestX();
        
        updateGbest = false;
    }
    
}




/*
 
 -------------------------------------------------------------------------------------------
                                Printing Methods
 -------------------------------------------------------------------------------------------
 
 Description:
    The below methods print various aspects of the swarm search to the console.
 
 */

void equivswarm::printGbest(){
    
    cout << "\n The maximum value of the sensitivity function identified is: " << gbest << endl;
    cout << "Corresponding leb: " << exp(-gbest / nParameter) << endl;
    
}

void equivswarm::printGbestX(){
    
    cout << "\n The candidate design point obtaining the maximum is: \n" << gbestX << endl;
    
}

bool equivswarm::checkConvergence(){
    
    bool isConverged = false;
    
    if ( exp(-gbest / nParameter) <= 1.0){
        if (minval / gbest > 0.9999){
            cout << "converged on iteration: " << iter ;
            isConverged = true;
            converged = true;
        }
    }
    
    return(isConverged);
    
}



