//
//  equivswarm.hpp
//  QPSO


#ifndef equivswarm_hpp
#define equivswarm_hpp

#include "qpso.h"

/*
 EQUIVSWARM
 
 -------------------------------------------------------------------------------------------
 Class will contain an object called eqSwarm. Each element of the swarm is a equivparticle object.
 -------------------------------------------------------------------------------------------
 
 Class Attributes:
 **eqSwarm      :    An array of pointers to qparticle objects
 invFisherInfo  :    The inverse fisher information. Used in calculation of sensitivity function
 interactions   :    Matrix where each row is an interaction
 beta           :    Vector of nominal parameter values
 lb             :    The lower limit for each design point. Note that this does not include the intercept.
 ub             :    The upper limit for each design point. Note that this does not include the intercept.
 gbestX         :    Vector of the design point found by the swarm with the largest returned sensitivity fn
 mbestX         :    Mean vector of each particle's pbestX
 nCont          :    int number of continuous factors
 nDis           :    int number of discrete factors
 nInt           :    int number of interactions
 nCol           :    int number of columns of the design matrix XXX does not currently appear to be used
 maxit          :    int maximum number of swarm iterations to try to find the largest value of sensitivity fn
 iter           :    int current iteration
 nPart          :    int number of particles in the swarm
 nParameter     :    int number of parameters in the model, should be the same as the length of beta
 gbestIndex     :    int index of which particle has the best known position
 worstIndex     :    int index of which particle has the worst known position
 updateGbest    :    bool is true if this iteration the gbest position increased
 converged      :    bool is true if the swarm has converged to a solution
 gbest          :    double keeps track of the largest value of the sensitivity function
 minval         :    double keeps track of the worst value of the sensitivity function
 alpha          :    double constriction rate XXX check how this is decremented
 fracPart       :    double equal to 1/nPart, just used in taking average for mbest
 */


class equivparticle; // Not sure why I needed this forward declaration, header file should have worked?

class equivswarm{
    
public:
    
    // Class attributes
    equivparticle **eqSwarm;
    MatrixXd invFisherInfo; MatrixXd interactions;
    VectorXd beta; VectorXd lb; VectorXd ub; VectorXd gbestX; VectorXd mbestX;
    int nCont, nDis, nInt, nCol, maxit, iter, nPart, nParameter;
    int gbestIndex = 0; int worstIndex = 0;
    bool updateGbest = false; bool converged = false;
    double gbest = -99999; double minval; double alpha; double fracPart;
    
    // Class Constructor and initializer
    equivswarm();
    void intitilizeEquivSwarm(MatrixXd fisherInformation, VectorXd b, int nContinuous, int nDiscrete, VectorXd lbd, VectorXd ubd, int nParticle, int maxiter, int nInteractions, MatrixXd interactionMat);
    
    // Updating Methods
    void updateMBest();
    void updateIteratorLocal();
    
    // Printing Methods
    void printGbest();
    void printGbestX();
    
    // Returning Methods
    bool checkConvergence();
    double getELB(){ return( exp(-gbest / nParameter ) ); }
    
    ~equivswarm();
    
};

#endif /* equivswarm_hpp */
