//
//  main.cpp
//  QPSO
//

#include "qpso.h"

/*
 
 MAIN.CPP
 
 Description:
    This file exists to demonstrate using d-QPSO to find designs for the odor removal experiment.
    Examples of finding locally D-optimal, Pseudo-Bayesian, and exact locally D-optimal designs are 
    provided. To run any specific example uncomment the code. The arguments to the "qpso" method are
    listed below (in order) for ease of reference.
 
 Arguments:
 num_swarm                   - integer number of quantum-behaved particle swarms to use
 num_parts                   - integer number of particles in each swarm
 maxit                       - integer maximum number of iterations
 maxresets                   - integer maximum number of times the swarm will reset if it
                                fails to find a design of the desired quality. Fix at 1
 lower_efficiency_bound      - lower bound on the relative efficiency of the obtained design
                                to the theoretically D-optimal design required to terminate.
                                Not used for bayesian approach.
 link                        - link function to relate linear predictor to outcome
 rowLimit                    - maximum number of support points. For exact designs this is N.
 priors                      - string vector of priors. Currently handles uniform "U(x,y)" and
                                Normal "N(x,y)" distributions. Simply use {"none"} if instead looking for locally optimal designs
 breedIterNum                - integer number of iterations between elitist breeding mutations
 outpath                     - filepath to the folder where the design is to be stored
 prefix                      - name for the csv file with the design
 logging                     - boolean, if 1, then pbests for each iteration are saved. Do not use
                                if using lots of particles or swarms (will create huge file)
 allocation                  - string: "exact" for exact designs, "approx" for approximate designs
 printIterBest               - if TRUE, each swarms gbest is printed at the end of each iteration
 printFinalInfo              - if TRUE, the final design is printed to the console
 saveFinalDesign             - if TRUE, the final design is saved to outpath/prefix_gbest_design.csv
 
 */



int main(int argc, const char * argv[]) {
    
    std::clock_t startcputime = std::clock();
    
    // Settings
     int nfactor = 5;
    
    // lower bound for each factor
     VectorXd lbound(nfactor);
     lbound(0) = -1; lbound(1) = -1; lbound(2) = -1; lbound(3) = -1; lbound(4) = 5;
    
    // upper bound for each factor
     VectorXd ubound(nfactor);
     ubound(0) = 1; ubound(1) = 1; ubound(2) = 1; ubound(3) = 1; ubound(4) = 35;
    
    // Nominal values for the betas (local optimality)
     VectorXd beta(nfactor + 1);
     beta(0) = -1.0; beta(1) = 2.0; beta(2) = 0.5; beta(3) = -1; beta(4) = -0.25; beta(5) = 0.13;
    
    // path to output folder
    std::string outpath = "/FILEPATH/TO/OUTPUT/FOLDER/";
    
    // Interactions (there arent any)
     VectorXd intact(1);
    
    designProblem dp;
    
    
     //************************* Odor Example, Locally D Optimal **************************\
    dp.initializeDesignProbem(4, 1, 0, beta, lbound, ubound, intact, "local", "grid", "D");
    dp.qpso(6, 15, 7000, 1, 0.99, "logit", 15, {"None"}, 1, outpath, "odor_local_approx", false, "approx", true, true, false);
    dp.fullSensitivityCheck(10000);
    
    
    
    //************************* Odor Example, Pseudo Bayesian ************************* \\
    
    //dp.initializeDesignProbem(4, 1, 0, beta, lbound, ubound, intact, "bayes", "grid", "D");
    
    // Prior is +- 100% of nominal value
    //dp.qpso(8, 30, 10000, 1, 0.99, "logit", 16, {"U(-2.0,0.00)", "U(0.0,4.00)", "U(0.00,1.00)", "U(-2.00,0.00)", "U(-0.50,0.00)" , "U(0.00,0.26)"}, 1, outpath, "odor_bayes", false, "approx", true, true, true);
    
    
    
    //************************* Odor Example, Exact Locally D Optimal ************************* \\
    
    //dp.initializeDesignProbem(4, 1, 0, beta, lbound, ubound, intact, "local", "grid", "D");
    // This is N = 100
    //dp.qpso(5, 30, 5000, 1, 0.99, "logit", 100, {"None"}, 1, outpath, "odor_local_exact", false, "exact", true, true, true);

    
    double cpu_duration = (std::clock() - startcputime) / (double)CLOCKS_PER_SEC;
    std::cout << "Finished in " << cpu_duration << " seconds [CPU Clock] " << std::endl;
    
    return 0;
}
