//
//  qhabitat.cpp
//  QPSO


#include "qhabitat.hpp"

qhabitat::qhabitat(){};
qhabitat::~qhabitat(){};

/*
 
 ----------------------------------------------------------------------------
                            initializeHabitat
 ----------------------------------------------------------------------------
 
 Arguments:
 numOfSwarms                   - integer number of particle swarms
 numParticle                   - integer number of particles in each swarm
 IAnew                         - integrationAbscissas object create for pseudo Bayesian optimality approach
 lbd                           - Eigen vector of lower bounds for each factor
 ubd                           - Eigen vector of upper bounds for each factor
 b                             - Eigen vector of nominal parameter values
 num_discrete                  - integer number of discrete factors
 num_continuous                - integer number of continuous factors
 num_interactions              - integer number of interactions
 interactions                  - Eigen matrix of interactions
 rowLimiti                     - integer maximum number of support points
 eb_criti                      - integer how often to do elitist breeding
 jump_num                      - integer number of positions to breed (Rec is 1)
 optTypeI                      - optimality Type (local or pseudo Bayesian)
 lower_bound_to_quit           - double minimum value for the lower efficiency bound to terminate early
 lowerLimit                    - Eigen vector of lower limits for EW designs, no longer used
 upperLimit                    - Eigen vector of upper limits for EW designs, no longer used
 loggingSetting                - boolean. If true, pbests are stored from each iteration to file
 assign                        - string that is "exact" for exact designs and "approx" for approximate designs
 sensType                      - string "grid" for grid search and "pso" for swarm search to check the sensitivity function
 printIterBest_set             - boolean, if true then at the end of each iteration the gbest in each swarm is printed
 optimalityCrit                - string of D or A optimality. Currently only D is implemented
 
 Description:
 This function initializes a qhabitat object, which contains all the swarms and particles and manages their search for the optimum design.
 
 */

void qhabitat::initializeHabitat(int numOfSwarms, int numParticle, int maxiter, integrationAbscissas* IAnew, VectorXd lbd, VectorXd ubd, VectorXd b, int num_discrete, int num_continuous, int num_interaction, MatrixXd interactions, int rowLimitI, int eb_criti, int jump_num, std::string optTypeI, double lower_bound_to_quit, VectorXd lowerLimit, VectorXd upperLimit, bool loggingSetting, std::string assign, std::string sensType, bool printIterBest_set, std::string optimalityCrit){
    
    // Resize and assign beta
    nBeta = (int)b.size();
    beta.resize( nBeta );
    beta = b;
    nCol = 1 + num_discrete + num_continuous;
    nDis = num_discrete;
    nCont = num_continuous;
    nInt = num_interaction;
    optType = optTypeI;
    rowLimit = rowLimitI;
    lower_efficiency_bound = lower_bound_to_quit;
    
    // A or D optimality
    optCrit = optimalityCrit;
    
    leb = -999.9;
    sensitivityType = sensType;
    
    // Exact or continuous design
    allocate = assign;
    
    printIterBest = printIterBest_set;
    
    // lower and upper bounds
    lb.resize( num_discrete + num_continuous );
    ub.resize( num_discrete + num_continuous );
    lb = lbd; ub = ubd;
    
    interactionMatrix.resize(num_interaction, 1 + num_discrete + num_continuous);
    interactionMatrix = interactions;
    
    nPart = numParticle;
    
    // Assign the number of swarms and the maximum number of swarms;
    nSwarm = numOfSwarms;
    
    // Assign the maximum number of iterations
    maxit = maxiter;
    
    // Resize the habitat best values
    hBestP.resize(rowLimit);
    hBestX.resize(rowLimit, 1 + num_discrete + num_continuous);
    
    // Loop through and initialize each swarm
    habitat = new qswarm *[nSwarm];
    
    // Initialize each of the quantum particle swarms
    for (int iSwarm = 0; iSwarm < nSwarm; iSwarm++){
        habitat[iSwarm] = new qswarm();
        habitat[iSwarm]->initilizeSwarm(numParticle, maxiter, IAnew, lbd, ubd, b, num_discrete, num_continuous, num_interaction, interactions, rowLimit, eb_criti, jump_num, optTypeI, lower_efficiency_bound, lowerLimit, upperLimit, loggingSetting, iSwarm, allocate, optCrit);
    }
    
}

/*
 
 ----------------------------------------------------------------------------
                            runEntireSearch
 ----------------------------------------------------------------------------

 Description:
    This is the key function in the habitat. It performs all iteration updates
    for all particle swarms. The steps the function takes depend on the type of
    optimal design of interest (approximate vs pseudo-Bayesian).
 
 */

void qhabitat::runEntireSearch(){
    
    /* could add this is for bayesian searches, currently not done in this location
    integrationAbscissas *IA = new integrationAbscissas;
    if (optType == "bayes"){
        IA->performPriorConversion(priors);
        IA->generateRadialQuantities(nParameter);
        IA->generateSimplexMethod(nParameter);
        IA->generateOrthoRotations(nParameter);
        IA->generateThetaSearchPoints(nParameter);
        IA->performPriorApproximation();
    }
     */
    
    bool converged = false;
    
    for (int iIter = 0; iIter < maxit; iIter++){
        
        cout << iIter << endl;
        
        for (int iSwarm = 0; iSwarm < nSwarm; iSwarm++){
            if (optType == "local"){
                    habitat[iSwarm]->updateMBest();
                    habitat[iSwarm]->updateIteratorLocal();
                    if (iIter > 50 && allocate == "approx"){habitat[iSwarm]->tweakProps();}
                    converged = habitat[iSwarm]->checkConvergence();
                    if (converged){iIter = maxit;}
                
                // check for an improvement to the habitat best
                if (habitat[iSwarm]->getGbest() > hBest){
                    hBest = habitat[iSwarm]->getGbest();
                    bestSwarm = iSwarm;
                    improved = true;
                }
                
            } else if (optType == "bayes"){
                    habitat[iSwarm]->updateMBest();
                    habitat[iSwarm]->updateIteratorPseudoBayes();
                    if (allocate == "approx"){habitat[iSwarm]->random_rounding();}
                    converged = habitat[iSwarm]->checkConvergence();
                    if (converged){iIter = maxit;}
                // check for an improvement to the habitat best
                if (habitat[iSwarm]->getGbest() > hBest){
                    hBest = habitat[iSwarm]->getGbest();
                    bestSwarm = iSwarm;
                }
            }
        }
        
        // Perform elitist breeding at a swarm level
        if (iIter % 2 == 0 & iIter > 200 & nSwarm > 1 & optType=="local"){
            elitist_breed_swarms();
            //crossoverSwarms();
        }
        
        if (iIter % 20 == 0 & iIter > (20*nBeta) & improved & optType == "local"){
            checkBoundSatisfied();
            improved = false;
            if (boundSatisfied){ iIter = maxit;}
        }
        
        // print fitness from each swarm
        if (printIterBest){
            for (int iSwarm = 0; iSwarm < nSwarm; iSwarm++){
                cout << habitat[iSwarm]->getGbest() << "   ";
            }
            cout << endl;
        }
        
    }
    
    // Save the best design
    hBestX = habitat[bestSwarm]->getGbestX();
    hBestP = habitat[bestSwarm]->getGbestP();

}


/*
 
 ----------------------------------------------------------------------------
                            getGbestFisherInformation
 ----------------------------------------------------------------------------
 
 Output:
    An eigen matrix object of the best gbest Fisher Information across all
    swarms. This is the hbest fisher information matrix
 
 Description:
    Function to return the information matrix corresponding to the best design
    across all particles and swarms.
 
 */

MatrixXd qhabitat::getGbestFisherInformation(){
    return habitat[bestSwarm]->getGbestFisherInformation();
}

/*
 
 ----------------------------------------------------------------------------
                            checkBoundSatisfied
 ----------------------------------------------------------------------------
 
 Description:
    Function to check if the lower bound on the efficiency of the hbest design
    is high enough to terminate.
 
*/

void qhabitat::checkBoundSatisfied(){
        
    double currentELB = 0;
    
    // First step is to check if the design is good enough at the design points
    if ( habitat[bestSwarm]->checkEquivalence() || overrideELB ){
        
        if (sensitivityType == "swarm"){
            // Setup a search for the elb using an equivalence swarm
            equivswarm checker;
            int maxEquivIt = 200;
            checker.intitilizeEquivSwarm( getGbestFisherInformation(), beta, nCont, nDis, lb, ub, 75, maxEquivIt, nInt, interactionMatrix);
            bool conv = false;
            for (int iter = 0; iter < maxEquivIt; iter++){
                checker.updateMBest();
                checker.updateIteratorLocal();
                conv = checker.checkConvergence();
                if (conv){iter = maxEquivIt;}
            }
            currentELB = checker.getELB();
        }
        
        // Check the sensitivity function using a grid search, not rec for > 3 cont factors
        if (sensitivityType == "grid"){
            currentELB = sensitivityCheck(100);
        }
        
    }
    
    if (currentELB > lower_efficiency_bound){
        cout << "bound was satisfied with: " << currentELB << endl;
        boundSatisfied = true;
    }
    leb = currentELB;
    
}


/*
 
 ----------------------------------------------------------------------------
                            incrementSetting
 ----------------------------------------------------------------------------
 
 Arguments:
    setting         -  current design point setting
    nSearchValue    - eigen vector of the number of settings to search for each factor
    iFactor         - integer current factor being incremented
 
 Output:
    newSetting      - the next design point for the sensitivity function
 
 Description:
    Recursive helper function for calculating the sensitivity function.
    Finds the correct design point at which to test the sensitivity function.

 
 */

std::vector<int> qhabitat::incrementSetting(std::vector<int> setting, std::vector<int> nSearchValue, int iFactor){
    
    // Handle case where we are at the end
    if ( setting[iFactor] == nSearchValue[iFactor]-1 ){
        if (iFactor > 0){
            return incrementSetting( setting, nSearchValue, iFactor - 1 );
        }
        else{
            //cout << "\n We have hit the end" << endl;
            setting[0] = -999;
            return(setting);
        }
    }
    // Increment
    else{
        std::vector<int> newSetting; newSetting.resize( setting.size() ); newSetting = setting;
        newSetting[iFactor] = setting[iFactor] + 1;
        // reset everything to the right
        if ( iFactor < nSearchValue.size()-1 ){
            for (int iAbove = iFactor+1; iAbove < nSearchValue.size(); iAbove++ ){
                newSetting[iAbove] = 0;
            }
        }
        return(newSetting);
    }
    
}

/*
 
 ----------------------------------------------------------------------------
                            getLEB
 ----------------------------------------------------------------------------

 Output:
        leb - the lower bound on the efficiency of the hbest design
 
 Description:
    Function to calculate and return the value of the lower bound
    on the efficiency of the hbest design. This function is called
    outside the regular sensitivity/bound check, in case the user 
    desires to check the sensitivity function outside of the prespecified
    times.
 
 */

double qhabitat::getLEB(){
    if (boundSatisfied == false){
        overrideELB = true;
        checkBoundSatisfied();
    }
    overrideELB = false;
    return leb;
}

/*
 
 ----------------------------------------------------------------------------
                        sensitivityCheck
 ----------------------------------------------------------------------------
 
 Arguments:
    resolution  -   integer number of different points to break each continuous
                    factor into for the sensitvity check.
 
 Output:
    leb - the lower bound on the efficiency of the hbest design
 
 Description:
    Function to calculate and return the value of the lower bound
    on the efficiency of the hbest design.
 
 */

double qhabitat::sensitivityCheck(int resolution){
    
    // Need the inverse of the fisher information matrix
    MatrixXd fisherInfo(nBeta, nBeta);
    MatrixXd invFisherInfo(nBeta, nBeta);

    fisherInfo = getGbestFisherInformation();
    invFisherInfo = fisherInfo.inverse();
    
    // Empty structures
    int nFactor = (int)lb.size();
    std::vector<int> nSearchValue(nFactor);
    
    // Fill out the number of search values. Will be 2 for discrete factors and RESOLUTION for continuous
    for (int iFactor = 0; iFactor < nFactor; iFactor++){
        if (iFactor < nDis){ nSearchValue[iFactor] = 2; }
        else {  nSearchValue[iFactor] = resolution; }
    }
    
    // Now create an array of the size of the search values, so 2 x 2 x ... x 2 x res x res x ... x res
    std::vector< std::vector<double> > candValues;
    candValues.resize(nFactor);
    // Loop through the array and populate it with the values
    int nPoints = 0;
    double incr = 0;
    for (int iFactor  = 0; iFactor < nFactor; iFactor++){
        nPoints = nSearchValue[iFactor];
        // Resize this part of the array to the correct size (nPoints)
        candValues[iFactor].resize(nPoints);
        // Populate for a discrete factor
        if (nPoints == 2){ candValues[iFactor][0] = -1.0; candValues[iFactor][1] = 1.0; }
        // Populate for a continuous factor
        else{
            // get increment size
            incr = ( ub(iFactor) - lb(iFactor) ) / (resolution-1);
            // loop through and add values to search over
            double lowestValue = lb[iFactor];
            for (int iGrid = 0; iGrid < resolution; iGrid++){
                candValues[iFactor][iGrid] = lowestValue + ( iGrid * incr + 0.0 );
            }
        }
    }
    
    // Now need to evaluate each of the prod( nSearchValue ) possible design points
    // initialize a vector to track current position
    std::vector<int> currentSetting; currentSetting.resize(nFactor);
    for (int ii = 0; ii < nFactor; ii++){ currentSetting[ii] = 0; }
    // Create vector to store current candidate design point and assign the intercept
    VectorXd xPoint (nBeta); xPoint(0) = 1;
    double wPoint = 0;
    double xbPoint = 0;
    double sensValue = 0;
    double maxSens = -9999;
    int nDesignColumn = (1 + nDis + nCont);
    // Obtain the current setting
    while (currentSetting[0] > -998){
        
        // set the design point to the current setting
        for (int iFactor = 1; iFactor < nDesignColumn; iFactor++){
            xPoint(iFactor) = candValues[iFactor-1][currentSetting[iFactor-1]];
        }
        
        // Loop over the interactions
        for (int iInteract = 0; iInteract < nInt; iInteract++){
            double interactionValue = 1.0;
            for (int iCol = 1; iCol < nDesignColumn; iCol++){
                //cout << interactionMatrix(iInt, iCol) * X(iRow, iCol) << "\n";
                if (interactionMatrix(iInteract, iCol) != 0){
                    interactionValue *= interactionMatrix(iInteract, iCol) * xPoint(iCol);
                }
            }
            xPoint(iInteract + nDesignColumn) = interactionValue;
        }
        
        // Calculate the value of w at this design point
        xbPoint = xPoint.transpose() * beta;
        wPoint = exp(xbPoint) / ( ( 1 + exp(xbPoint) ) * ( 1+exp(xbPoint) ) );
        
        // Evaluate the sensitivity function
        sensValue = wPoint * xPoint.transpose() * invFisherInfo * xPoint - nBeta;
        
        if (sensValue > maxSens){maxSens = sensValue;}
        
        // Obtain the current setting for the next point
        currentSetting = incrementSetting(currentSetting, nSearchValue, nFactor-1);
        
        // This function only runs to check if the sensitvity function is satisfied, not to provide the leb, therefore it should terminante if the leb is not satisfied.
        if ( exp(-maxSens / nBeta) < lower_efficiency_bound && overrideELB == false ){
            currentSetting[0] = -998;
        }
        
    }
    
    leb = exp(-maxSens / nBeta);
    
    if (leb > lowestBoundAchieved){
        lowestBoundAchieved = leb;
    }
    
    return(leb);
}


/*
 
 ----------------------------------------------------------------------------
                        elitist_breed_swarms
 ----------------------------------------------------------------------------
 
 Description:
    Function to handle elitist breeding between two swarms.
 
 */

void qhabitat::elitist_breed_swarms(){
    
    std::random_device rdm;
    std::mt19937 generator(rdm());
    std::uniform_real_distribution<double> pdist(0,1);
    
    // Initialize matrices
    MatrixXd epool_i(rowLimit, nCol);
    VectorXd epoolP_i(rowLimit);
    MatrixXd epool_compare(rowLimit, nCol);
    VectorXd epoolP_compare(rowLimit);
    MatrixXd normedX (rowLimit, nCol);
    
    double tspn = 0.0; double newFitness = -999.0;
    
    bool EBoccured = false;
    bool propoccured = false;
    
    int partIndex = 0;
    int swarmIndex = 0;
    int iCompare = 0;
    
    int tspnIndexRow = 0;
    int tspnIndexCol = 0;
    int tspnIndexRowMove = 0;
    int tspnIndexColMove = 0;
    
    int iRowCurrent = 0; int iColCurrent = 0;
    int iRowCurrentMove = 0; int iColCurrentMove = 0;

    int jumpNum = 1;
    std::vector< std::vector<int> > iTspn( jumpNum, std::vector<int> (2)  );
    std::vector< std::vector<int> > iTspnMove( jumpNum, std::vector<int> (2)  );

    for (int iSwarm = 0; iSwarm < nSwarm; iSwarm++){
        
        // Change of this swarm undergoing breeding
        if (pdist(generator) < 0.25){
            
            // Loop over each particle
            for (int iPart = 0; iPart < nPart; iPart++){
                
                EBoccured = false; propoccured = false;
                newFitness = -999;
                
                // probability of this particle being bred
                if (pdist(generator) < 0.5){
                    
                    // select another swarm to breed with
                    swarmIndex = ceil( (pdist(generator) * (nSwarm)) - 1 );
                    while (swarmIndex == iSwarm){
                        swarmIndex = ceil( (pdist(generator) * (nSwarm)) - 1 );
                    }
                    // select another particle to breed with
                    partIndex = ceil( (pdist(generator) * (nPart)) - 1 );
                    
                    // PERFORM ELITIST BREEDING - DESIGN POINT
                    if (pdist(generator) < 0.1){
                        EBoccured = true;
                        epoolP_i = habitat[iSwarm]->getPbestP(iPart);
                        iCompare = ceil( (pdist(generator) * (nPart)) - 1 );
                        tspn = pdist(generator);
                        // Get the edited index start point; create an index list, these act as starting pt
                        tspnIndexRow = floor( (pdist(generator) * (rowLimit)) );
                        tspnIndexCol = ceil( (pdist(generator) * (nCol-1)) );
                        // Copy the index list; these will be incremented
                        iRowCurrent = tspnIndexRow; iColCurrent = tspnIndexCol;
                        // Loop and increment
                        for (int iRow = 0; iRow < jumpNum; iRow++){
                            iTspn[iRow][0] = iRowCurrent;
                            iTspn[iRow][1] = iColCurrent;
                            iColCurrent++;
                            if (iColCurrent == nCol){
                                iColCurrent = 1;
                                iRowCurrent++;
                            }
                            if (iRowCurrent == rowLimit){
                                iRowCurrent = 0;
                            }
                        }
                        // Those indices are to be moved somewhere else. Need to create that index list as well;
                        tspnIndexRowMove = floor( (pdist(generator) * (rowLimit)) );
                        tspnIndexColMove = ceil( (pdist(generator) * (nCol-1)) );
                        // Copy the index list; these will be incremented
                        iRowCurrentMove = tspnIndexRowMove; iColCurrentMove = tspnIndexColMove;
                        // Loop and increment
                        for (int iRow = 0; iRow < jumpNum; iRow++){
                            iTspnMove[iRow][0] = iRowCurrentMove;
                            iTspnMove[iRow][1] = iColCurrentMove;
                            iColCurrentMove++;
                            if (iColCurrentMove == nCol){
                                iColCurrentMove = 1;
                                iRowCurrentMove++;
                            }
                            if (iRowCurrentMove == rowLimit){
                                iRowCurrentMove = 0;
                            }
                        }
                        
                        // Handle operations between particle copies
                        normedX = normalizeDesignEB( habitat[iSwarm]->getPbestX(iPart) );
                        epool_i = normedX;
                        epool_compare = normalizeDesignEB( habitat[swarmIndex]->getPbestX(iCompare) );

                        //cout << "designw was \n" << swarm[iPart]->getPbestX();
                        for (int iRow = 0; iRow < jumpNum; iRow++){
                            epool_i(iTspnMove[iRow][0], iTspnMove[iRow][1]) = epool_compare( iTspn[iRow][0], iTspn[iRow][1]) ;
                        }
                            
                        epool_i = revNormalizeDesignEB( epool_compare );
                        
                    } // end elistist breeding - design point
                    
                    // Elitist breeding - proportions goes here
                    if (pdist(generator) < 0.9 && allocate == "approx"){
                        
                        
                        propoccured = true;
                        VectorXd epoolP_i_compare(rowLimit);
                        
                        // If the first EB did not happen, then set epool_i to X
                        if (!EBoccured){
                            epool_i = habitat[iSwarm]->getPbestX(iPart);
                        }
                        
                        EBoccured = true;
                        epoolP_i = habitat[iSwarm]->getPbestP(iPart);
                        iCompare = ceil( (pdist(generator) * (nPart)) - 1 );
                        tspn = pdist(generator);
                        
                        // Get the edited index start point; create an index list, these act as starting pt
                        std::vector<int> iTspn( jumpNum );
                        int tspnIndexRow = floor( (pdist(generator) * (rowLimit)) );
                        // Copy the index list; these will be incremented
                        int iRowCurrent = tspnIndexRow;
                        // Loop and increment
                        for (int iRow = 0; iRow < jumpNum; iRow++){
                            iTspn[iRow] = iRowCurrent;
                            iRowCurrent++;
                            if (iRowCurrent == rowLimit){
                                iRowCurrent = 0;
                            }
                        }
                        
                        std::vector<int> iTspnMove( jumpNum );
                        int tspnIndexRowMove = floor( (pdist(generator) * (rowLimit)) );
                        // Copy the index list; these will be incremented
                        int iRowCurrentMove = tspnIndexRowMove;
                        // Loop and increment
                        for (int iRow = 0; iRow < jumpNum; iRow++){
                            iTspnMove[iRow] = iRowCurrentMove;
                            iRowCurrentMove++;
                            if (iRowCurrentMove == rowLimit){
                                iRowCurrentMove = 0;
                            }
                        }
                        
                        epoolP_i = habitat[iSwarm]->getPbestP(iPart);
                        epoolP_i_compare = habitat[swarmIndex]->getPbestP(iCompare);
                        
                        // Perform the swap
                        for (int iRow = 0; iRow < jumpNum; iRow++){
                            epoolP_i(iTspnMove[iRow]) = epoolP_i_compare( iTspn[iRow]) ;
                        }
                        
                        // Make sum to 1
                        double sumP = epoolP_i.sum();
                        for (int iRow = 0; iRow < rowLimit; iRow++){
                            epoolP_i(iRow) = epoolP_i(iRow) / sumP;
                        }
                        
                    } // end proportion breeding
                    
                    
                } // end breeding for this particle
                
                // IF breeding occured need to check for improvement
                if (EBoccured){
                    // Check the fitness of the EB particle.
                    if (optType == "local"){
                        newFitness = check_epool_fitness(epool_i, epoolP_i);
                    } else if (optType == "bayes"){
                        //newFitness = check_epool_fitness_pseudoBayes(epool_i, epoolP_i);
                    }
                    

                    
                    if (newFitness > habitat[iSwarm]->getPbest(iPart)){
                        habitat[iSwarm]->setX(iPart, epool_i);
                        habitat[iSwarm]->setP(iPart, epoolP_i);
                        if (allocate == "approx"){
                            habitat[iSwarm]->checkDuplicatePoints(iPart);
                        }
                        habitat[iSwarm]->frameArrange(iPart);
                        habitat[iSwarm]->setPbestX( iPart, habitat[iSwarm]->getX(iPart) );
                        habitat[iSwarm]->setPbestP( iPart, habitat[iSwarm]->getP(iPart) );
                        habitat[iSwarm]->setPbest(iPart, newFitness);
                        if ( newFitness > habitat[iSwarm]->getGbest() ){
                            habitat[iSwarm]->setGbestDirect( newFitness );
                            habitat[iSwarm]->setGbestX( habitat[iSwarm]->getX(iPart) );
                            habitat[iSwarm]->setGbestP( habitat[iSwarm]->getP(iPart) );
                            habitat[iSwarm]->setGbestIndex( iPart );
                            habitat[iSwarm]->setGbestDirect( newFitness);
                        } // end check if new fitness is better than gbest
                        
                    } // end check if new fitness is better than pbest
                    
                } // end check if EB occured
                
            } // end of loop over each particle
            
            
        } // end of check if swarm should be bred
        
        
        
    }
    
}



/*
 
 ----------------------------------------------------------------------------
                        normalizeDesignEB
 ----------------------------------------------------------------------------
 
 Arguments:
    X                  - Eigen matrix object, is an experimental design
 
 Output:
    normDesign         - Eigen matrix object, normalized experimental design
 
 Description:
    The elitist breeding mutator pastes factor levels onto other factor levels
    the factor levels do not alwaays have the same range, thus the conversion is
    performed on the normalized (0,1) scale and then backtransformed to the true
    factor range using the revNormalizeDesignEB function.
 
 */

MatrixXd qhabitat::normalizeDesignEB(MatrixXd X){
    MatrixXd normDesign (rowLimit, nCol);
    
    for (int iRow = 0; iRow < rowLimit; iRow++){
        normDesign(iRow, 0) = 1;
        for (int iCol = 1; iCol < nCol; iCol++){
            normDesign(iRow, iCol) = ( X(iRow, iCol) - lb(iCol-1) ) / ( ub(iCol - 1) - lb(iCol - 1) );
        }
    }
    
    return(normDesign);
}

/*
 
 ----------------------------------------------------------------------------
                            revNormalizeDesignEB
 ----------------------------------------------------------------------------
 
 Arguments:
    X                  - Eigen matrix object, normalized experimental design
 
 Output:
    revNormDesign      - Eigen matrix object, experimental design in original scale
 
 Description:
    Function that takes a normalized experimental design (see normalizeDesignEB function)
    and converts it back to the original factor ranges.
 
 */

MatrixXd qhabitat::revNormalizeDesignEB(MatrixXd X){
    MatrixXd revNormDesign (rowLimit, nCol);
    
    // Reverse the normalization
    for (int iRow = 0; iRow < rowLimit; iRow++){
        revNormDesign(iRow, 0) = 1.0;
        // Correct the discrete factors
        for (int iCol = 1; iCol < nDis+1; iCol++){
            if (X(iRow, iCol) < 0.5){
                revNormDesign(iRow, iCol) = lb(iCol-1);
            } else{
                revNormDesign(iRow, iCol) = ub(iCol-1);
            }
        }
        // Correct the continuous factors
        for (int iCol = nDis+1; iCol < nCol; iCol++){
            revNormDesign(iRow, iCol) = X(iRow, iCol)*(ub(iCol-1)-lb(iCol-1)) + lb(iCol-1) ;
        }
    }
    
    return(revNormDesign);
}

/*
 
 ----------------------------------------------------------------------------
                            check_epool_fitness
 ----------------------------------------------------------------------------
 
 Arguments:
    epool        - Eigen matrix object, candidate experimental design from 
                    elitist breeding
    eP           - Eigen vector object of the elitist breeding proportions
 
 Output:
    fitness      - The fitness of the elitist breeding design
 
 Description:
    Function to check the fitness of the elitist breeding design.
 
 */

double qhabitat::check_epool_fitness(MatrixXd epool, VectorXd eP){
    double fitness = 0;
    
    // Initialize the W matrix to be all zeros
    MatrixXd W (rowLimit, rowLimit);
    for (int iRow = 0; iRow < rowLimit; iRow++){
        for (int iCol = 0; iCol < rowLimit; iCol++){
            W(iRow, iCol) = 0.0;
        }
    }
    
    // Get the full design point including interaction
    
    MatrixXd fullX(rowLimit, nCol + nInt);
    // Calculate the interactions
    MatrixXd interactionBlock(rowLimit, nBeta - nCol);
    for (int iRow = 0; iRow < rowLimit; iRow++){
        for (int iInt = 0; iInt < nInt; iInt++){
            double interactionValue = 1.0;
            for (int iCol = 1; iCol < nCol; iCol++){
                //cout << interactionMatrix(iInt, iCol) * X(iRow, iCol) << "\n";
                if (interactionMatrix(iInt, iCol) == 1){
                    interactionValue *= interactionMatrix(iInt, iCol) * epool(iRow, iCol);
                }
            }
            interactionBlock(iRow, iInt) = interactionValue;
        }
    }
    fullX << epool, interactionBlock;
    
    // Calculate the diagonal elements of the W matrix
    for (int iRow = 0; iRow < rowLimit; iRow++){
        double xb = fullX.row(iRow) * beta;
        W(iRow, iRow) = eP(iRow) * exp(xb) / ( (1+exp(xb)) * (1+exp(xb)) );
    }
    
    // Calculate the fisher information matrix
    MatrixXd info (nBeta, nBeta);
    info = fullX.transpose() * W * fullX;
    
    fitness = log(info.determinant());
    
    return(fitness);
}

/*
 
 ----------------------------------------------------------------------------
                            saveHBestDesign
 ----------------------------------------------------------------------------
 
 Arguments:
    filepath    - string path to folder where the design should be saved
    filename    - string filename for the hbest design
 
 Description:
    Writes the hbest design to a csv file.
 
 */

void qhabitat::saveHBestDesign(std::string outpath, std::string prefix){
    habitat[bestSwarm]->saveGbestDesign(outpath, prefix);
}

/*
 
 ----------------------------------------------------------------------------
                        printHBestDesign
 ----------------------------------------------------------------------------
 
 Description:
    Function to print the hbest design to the console.
 
 */

void qhabitat::printHBestDesign(){
    
    for (int iRow = 0; iRow < rowLimit; iRow++){
        if (hBestP(iRow) > 0.000001){
            for (int iCol = 0; iCol < nCol; iCol++){
                cout << hBestX(iRow, iCol) << " ";
            }
            cout << hBestP(iRow) << endl;
        }
    }
    
};





