//
//  qhabitat.hpp
//  QPSO


/*
 
 QHABITAT
 
 -------------------------------------------------------------------------------------------
                Class that manages all the individual particle swarms
 -------------------------------------------------------------------------------------------
 
 Class Attributes:
    habitat             - array of qswarm pointers
    hbestX              - Eigen matrix object of the best design found by any swarm/particle
    interactionMatrix   - Eigen matrix object where each row is an interaction
    hBestP              - Eigen vector object containing the best proportions for each support point
    beta                - Eigen vector object containing the nominal values for the parameters
    lb                  - Eigen vector object containing the lower bound for each factor
    ub                  - Eigen vector object containing the upper bound for each factor
    nCont               - integer number of continuous factors in the model
    nDis                - integer number of discrete factors in the model
    nInt                - integer number of interactions in the model
    rowLimit            - integer maximum number of support points
    nCol                - integer number of columns in the design matrix
    maxit               - integer maximum number of iterations
    nBeta               - integer number of parameters in the model
    nPart               - integer number of particles per swarm
    nSwarm              - integer number of particle swarms
    bestSwarm           - integer index of the best swarm
    hBest               - double best fitness value found among all swarms
    leb                 - lower bound on the relative efficiency of hbest design relative to d-opt design
    lowestBoundAchieved - best leb value found
    lower_efficiency_bound - leb required to terminate the search
    boundSatisfied      - boolean, true if the leb is high enough, results in termination of search
    sensitivityCheckSatisfied - boolean, true if the leb is high enough, results in termination of search
    improved            - boolean, true if the hbest has improved since the last sensitivity fn check
    overrideELB         - boolean, if true then the sensitivity function is checked regardless of status
    printInterBest      - boolean, if true then gbest for each swarm is printed to console at the end of each iter
    allocation          - string "approx" or "exact" for design point allocation
    optCrit             - string "D" or "A" optimality. Currently only D is fully implemented
    optType             - string "local" vs "bayes"
    sensitivityType     - string "pso" vs "grid" search for the maximum of the sensitivity function
 
 */

#ifndef qhabitat_hpp
#define qhabitat_hpp

#include "qpso.h"

class qswarm;

class qhabitat{
    
public:
    
    qswarm **habitat;
    
    MatrixXd interactionMatrix, hBestX;
    
    VectorXd beta, hBestP, lb, ub;
    
    int nSwarm, maxit, bestSwarm = 0; int rowLimit, nCol;
    int nPart, nInt, nDis, nCont, nBeta;
    
    double hBest = -9999999; double leb = -999.9; double lowestBoundAchieved = 0.0;
    double lower_efficiency_bound = 1.0;
    
    bool boundSatisfied = false;
    bool sensitivityCheckSatisfied = false;
    bool improved = false;
    bool overrideELB = false;
    bool printIterBest = false;
    
    std::string allocate = "approx";
    std::string sensitivityType;
    std::string optCrit;
    std::string optType;
    
    // Class Constructor and initializer
    qhabitat();
    void initializeHabitat(int numOfSwarms, int numParticle, int maxiter, integrationAbscissas* IAnew, VectorXd lbd, VectorXd ubd, VectorXd b, int num_discrete, int num_continuous, int num_interaction, MatrixXd interactions, int rowLimit, int eb_criti, int jump_num, std::string optTypeI, double lower_efficiency_bound, VectorXd lowerLimit, VectorXd upperLimit, bool loggingSetting, std::string assignment, std::string sensitivtyType, bool printIterBest_set, std::string optimalityCrit);
    
    // Master Loop Function
    void runEntireSearch();
    
    // Mutators
    void elitist_breed_swarms();
    MatrixXd normalizeDesignEB(MatrixXd X);
    MatrixXd revNormalizeDesignEB(MatrixXd X);
    double check_epool_fitness(MatrixXd epool, VectorXd eP);

    // Updater Functions
    void runIteration();
    void checkBoundSatisfied();
    double sensitivityCheck(int resolution);
    std::vector<int> incrementSetting(std::vector<int> setting, std::vector<int> nSearchValue, int iFactor);
    
    // Getter Functions
    double getHBest(){ return hBest; }
    MatrixXd getHBestX(){ return hBestX; }
    VectorXd getHBestP(){ return hBestP; }
    MatrixXd getGbestFisherInformation();
    double getLEB();
    
    // Saving functions
    void saveHBestDesign(std::string outpath, std::string prefix);
    
    // Printing Functions
    void printHBestDesign();
    
    ~qhabitat();
    
};


#endif /* qhabitat_hpp */
