//
//  qparticle.cpp
//  QPSO



#include "qpso.h"

qparticle::qparticle(){
    
}

/*
 
 ----------------------------------------------------------------------------
                            initParticle
 ----------------------------------------------------------------------------
 
 Arguments:
    lbd             - Eigen vector object of lower bounds for each factor
    ubd             - Eigen vector object of upper bounds for each factor
    betaVec         - Eigen vector object of nominal parameter values
    num_discrete    - integer number of discrete factors
    num_continuous  - integer number of continuous factors
    num_interactions- integer number of interactions
    interactions    - Eigen matrix object specifying the interactions
    rl              -   integer number of support points
    maxiti          - integer maximum number of iterations
    lowerBoundBeta  -  lower bounds for the parameters (no longer used)
    upperBoundBeta  -  upper bounds for the parameters (no longer used)
    allocate        - string type of allocation: exact or approximate
    optimalityCrit  -   string optimality criterion. A- or D- optimality. Currently
                        only D is fully implemented.
 
 Description:
    This function generates a particle for the QPSO search with a random starting
    position conforming to the arguments given to the function.
 
 */


//******************** Initialization Function ********************//
void qparticle::initParticle(VectorXd lbd, VectorXd ubd, VectorXd betaVec, int num_discrete, int num_continuous, int num_interactions, MatrixXd interactions, int rl, int maxiti, VectorXd lowerBoundBeta, VectorXd upperBoundBeta, std::string allocate, std::string optimalityCrit){
    
    // Initialize pbest to a very low value
    pbest = -999999999999; fitness = -999999999999;
    
    // Exact vs continuous design
    allocation = allocate;
    
    // Store the factor type info
    nCont = num_continuous; nDis = num_discrete;
    nInteraction = num_interactions;
    
    // A or D optimality
    optCrit = optimalityCrit;
    
    // This is the 2^k frame
    frameSize = pow(2, (num_continuous + num_discrete) );
    
    maxit = maxiti;
    
    // Store the maximum number of rows and the
    rowLimit = rl; nCol = (int)ubd.size() + 1; // adding 1 for the interaction
    
    // Information needed to do EW designs
    nBeta = (int)betaVec.size();
    
    // resize the class members to the proper size
    lb.resize( lbd.size() );
    ub.resize( ubd.size() );
    beta.resize( betaVec.size() );
    X.resize(rowLimit, nCol); pbestX.resize(rowLimit, nCol);
    interactionMatrix.resize(nInteraction, nCol);
    fullX.resize(rowLimit, nBeta);
    localAtt.resize(rowLimit, nCol);
    fisher.resize( nCol, nCol );
    W.resize(rowLimit, rowLimit);
    P.resize(rowLimit); pbestP.resize(rowLimit); localAttP.resize(rowLimit);
    currentXRow.resize(nCol);
    
    interactionMatrix = interactions;
    
    // Make W a matrix of zeros;
    for (int iRow = 0; iRow < rowLimit; iRow++){
        for (int iCol = 0; iCol < rowLimit; iCol++){
            W(iRow, iCol) = 0;
        }
    }
    
    // Store lower, upper bound and the parameter vector
    lb = lbd;
    ub = ubd;
    beta = betaVec;
    
    // Initialize the initial plimit, the final pilimit, and the current limit
    plim = 1.0 / rowLimit;
    pliminit = 1.0 / rowLimit;
    plimfinal = 1.0 / nCol;
    pmin = pliminit;
    plimIncr = (plimfinal - pliminit) / (maxit/10.0);
    //cout << "plimincr is " << plimIncr << endl;
    pminIncr = ( pmin - 0 ) / (maxit/ 2.0 );
    
    
    // Get the frame together to reordering rows
    generateFrameArrangement();
    
    generateInitialLocation();
    
}

/*
 
 ----------------------------------------------------------------------------
                            generateInitialLocation
 ----------------------------------------------------------------------------
 
 Description:
    This is a helper function for particle initialization. This function
    carries out the actual position assignment at random for the particle.
 
 */

void qparticle::generateInitialLocation(){
    
    std::random_device rdm;
    std::mt19937 generator(rdm());
    std::uniform_real_distribution<double> pdist(0,1);
    std::uniform_real_distribution<double> pLimdist(0,pliminit);
    
    // fill out the intercept
    for (int i=0; i < rowLimit; i++){X(i,0) = 1; pbestX(i, 0) = 1;}
    
    // fill out the discrete factors
    for (int iRow = 0; iRow < rowLimit; iRow++){
        for (int iFact = 1; iFact < 1+nDis; iFact++){
            double randVal = pdist(generator);
            // Create the actual design based on the random draw. If rand
            if ( randVal > 0.5 ){ X(iRow, iFact) = 1;}
            else { X(iRow, iFact) = -1;}
        }
    }
    
    // fill out the continuous factors
    for (int iRow = 0; iRow < rowLimit; iRow++){
        for (int iCol = nDis+1; iCol < nCol; iCol++){
            std::uniform_real_distribution<double> distribution( lb(iCol-1), ub(iCol-1) );
            X(iRow, iCol) = distribution(generator);
        }
    }
    
    // Fill out proportions allocated to each design point
    if (allocation == "approx"){
        double sumP = 0;
        for (int iRow = 0; iRow < rowLimit; iRow++){
            P(iRow) = pLimdist(generator);
            if (P(iRow) < pmin){ P(iRow) = pmin; }
            sumP = sumP + abs(P(iRow));
        }
        
        for (int iRow = 0; iRow < rowLimit; iRow++){
            P(iRow) = abs( P(iRow) ) / sumP;
        }
        
    } else {
        for (int iRow = 0; iRow < rowLimit; iRow++){
            P(iRow) = 1.0;
        }
    }
    
    if (allocation == "approx"){
        checkDuplicatePoints();
        //cout << "after checking duplication we have: \n" << X;
    }
    
    // Now that the frame has been arranged, update the personal best information
    pbestP = P;
    pbestX = X;
    
}

/*
 
 ----------------------------------------------------------------------------
                            REinitParticle
 ----------------------------------------------------------------------------
 
 Description:
    Function to restart a particle. Not currently in use, but useful
    when considering possible schemes to increase swarm diversity.
 
 */


void qparticle::REinitParticle(){
    
    std::random_device rdm;
    std::mt19937 generator(rdm());
    std::uniform_real_distribution<double> pdist(0,1);
    std::uniform_real_distribution<double> pLimdist(0,plim);
    
    // fill out the intercept
    for (int i=0; i < rowLimit; i++){X(i,0) = 1; pbestX(i, 0) = 1;}
    
    // fill out the qbits and the discrete factors
    for (int iRow = 0; iRow < rowLimit; iRow++){
        qbit(iRow, 0) = 0; pbestQbit(iRow, 0) = 0;
        for (int iFact = 1; iFact < 1+nDis; iFact++){
            qbit(iRow, iFact) = pdist(generator); pbestQbit(iRow, iFact) = qbit(iRow, iFact);
            double randVal = pdist(generator);
            // Create the actual design based on the random draw. If rand
            if ( randVal > qbit(iRow, iFact) ){ X(iRow, iFact) = 1;}
            else { X(iRow, iFact) = -1;}
        }
    }
    
    // fill out the continuous factors
    for (int iRow = 0; iRow < rowLimit; iRow++){
        for (int iCol = nDis+1; iCol < nCol; iCol++){
            std::uniform_real_distribution<double> distribution( lb(iCol-1), ub(iCol-1) );
            X(iRow, iCol) = distribution(generator);
        }
    }
    
    if (allocation == "approx"){
    // Fill out proportions allocated to each design point
    double sumP = 0;
    for (int iRow = 0; iRow < rowLimit; iRow++){
        P(iRow) = pLimdist(generator);
        sumP = sumP + abs(P(iRow));
    }
    for (int iRow = 0; iRow < rowLimit; iRow++){
        P(iRow) = abs( P(iRow) ) / sumP;
        //pbestP(iRow) = P(iRow);
    }
    
        checkDuplicatePoints();
    }
    
    // Now that the frame has been arranged, update the personal best information
    pbestP = P;
    pbestX = X;
    
    pbest = -999999;
    
}

/*
 
 ----------------------------------------------------------------------------
                            reDrawProportions
 ----------------------------------------------------------------------------
 
 Description:
    This function leaves the factor settings for a particle the same, but
    redraws new random proportions for each support point. Not currently in use
    but helpful for attempting to increase swarm diversity.
 
 */

void qparticle::reDrawProportions(){
    
    /*
    std::random_device rdm;
    std::mt19937 generator(rdm());
    std::uniform_real_distribution<double> pdist(0,1);
    std::uniform_real_distribution<double> pLimdist(0,plim);
    
    if (allocation == "approx"){
    // Fill out proportions allocated to each design point
    double sumP = 0;
    for (int iRow = 0; iRow < rowLimit; iRow++){
        P(iRow) = pLimdist(generator);
        sumP = sumP + abs(P(iRow));
    }
    for (int iRow = 0; iRow < rowLimit; iRow++){
        P(iRow) = abs( P(iRow) ) / sumP;
    }
    
        checkDuplicatePoints();
    }
    
    //frameArrange();
    
    // Now that the frame has been arranged, update the personal best information
    pbestP = P;
    //pbestX = X;
    
    pbest = -999999;
     */
    
}

/*
 
 ----------------------------------------------------------------------------
                        reDrawDesignPoint
 ----------------------------------------------------------------------------
 
 Arguments:
    ind     - integer index of the design point to be replaces
    indSet  - set of possible support point indices that are already in use
 
 Description:
    This function takes the support point indexed by the ind argument and 
    replaces is with a support point not currently present in the design.
    This is not currently in use, but it useful for increasing swarm diversity.
 
 */

void qparticle::reDrawDesignPoint( std::vector<int> ind, std::vector<int> indSet){
    
    std::random_device rdm;
    std::mt19937 generator(rdm());
    std::uniform_real_distribution<double> pdist(0,1);
    
    int nRedraw = (int)ind.size();
    int nAvailableFrames = (int) indSet.size();
    int drawIndex = 0, frameToUse = 0;
    int XtoEdit = 0;
    
    double scaledLimit = 0;
    
    for (int iPoint  = 0; iPoint < nRedraw; iPoint++){
        // draw the frame index to use
        drawIndex = floor( pdist(generator) * nAvailableFrames );
        frameToUse = indSet[drawIndex];
        XtoEdit = ind[iPoint];
        // Now we need to regenerate a design point and make it fit in frame.
        // For the discrete factors this will just be the frame itself
        for (int iCol = 1; iCol <= nDis; iCol++){
            X(XtoEdit, iCol) = frameArrangement(frameToUse, iCol-1);
        }
        // Now need to update the continuous factors
        for (int iCol = nDis + 1; iCol < nCol; iCol++){
            scaledLimit = (  lb(iCol-1) + ub(iCol-1) ) / 2.0;
            if ( frameArrangement(frameToUse, iCol-1) < 0 ){
                std::uniform_real_distribution<double> distribution( lb(iCol-1), scaledLimit );
                X(XtoEdit, iCol) = distribution(generator);
            }
            else {
                std::uniform_real_distribution<double> distribution( scaledLimit, ub(iCol-1));
                X(XtoEdit, iCol) = distribution(generator);
            }
        }
    }
    
    
}

/*
 
 ----------------------------------------------------------------------------
                        generateFrameArrangement
 ----------------------------------------------------------------------------
 
 Description:
    This function generates a frame of the possible discrete combinations
    of support points as well as the possible signs of the continuous factors.
    The grid will be of size 2^k where k is the number of factors in the model.
    This frame is later used to reorganize the order of the support points in
    the design X for easy comparison.
 
 */

void qparticle::generateFrameArrangement(){
    
    frameArrangement.resize( frameSize, nCol - 1 );
    
    // fill out the boundframe; only going to keep the ones corresponding to discrete factors
    for (int col = 0; col < nCol-1; col++){
        int flipcount = pow(2, (nCol - col - 2));
        bool lowerbound = true;
        int counter = 0;
        for (int row =0; row < frameSize; row++){
            if (lowerbound == true){
                frameArrangement(row,col) = -1;
            }
            if (lowerbound == false){
                frameArrangement(row,col) = 1;
            }
            counter = counter + 1;
            if (counter == flipcount){
                counter = 0;
                lowerbound = !lowerbound;
            }
        }
    }

}

/*
 
 ----------------------------------------------------------------------------
                            getUnusedFrame
 ----------------------------------------------------------------------------
 
 Output:
    isUnusedSetting -   Eigen vector object of 1's where a suppot point combination
                        setting is not used by the design X.
 
 Description:
    This function checks each suppot point in the design X against the arrangement
    of possible support points (the frame generated by generateFrameArrangement) and returns
    a vector of unused support points.
 
 */

VectorXi qparticle::getUnusedFrame(){
    MatrixXd Xscaled(rowLimit, nCol);
    Xscaled = X;
    for (int iRow = 0; iRow < rowLimit; iRow++){
        for (int iCol = nDis+1; iCol < nCol; iCol++){
            Xscaled(iRow, iCol) = 2*(Xscaled(iRow, iCol) - lb(iCol-1)) / ( ub(iCol-1) - lb(iCol-1) ) - 1;
        }
    }
    
    MatrixXi signs(rowLimit, nCol-1);
    for (int iRow = 0; iRow < rowLimit; iRow++){
        for (int iCol = 0; iCol < nCol-1; iCol++){
            signs(iRow, iCol) = 2 * (Xscaled(iRow, iCol+1) > 0) - 1;
        }
    }
    
    VectorXi belongsTo(rowLimit);
    
    // Now create a vector that is 1 if a row in the frame is unused, 0 otherwise
    VectorXi isUnusedSetting( frameSize );
    for (int iRow = 0; iRow < frameSize; iRow++){ isUnusedSetting(iRow) = 1; }
    
    int compareRow = 0;
    for (int iRow = 0; iRow < rowLimit; iRow++){
        
        // Get the sign of the first value; use it to pick a starting point
        if (signs(iRow, 0) > 0){ compareRow = frameSize / 2; } else {compareRow = 0;}
        
        for (compareRow; compareRow < frameSize; compareRow++){
            bool isSame = true;
            
            for (int iCol = 0; iCol < nDis; iCol++){
                
                if ( signs(iRow, iCol) != frameArrangement(compareRow, iCol) ){ iCol = nCol; isSame = false;}
                
            } // end column loop
            
            // If no differences were found, terminate search for this row;
            if (isSame){belongsTo(iRow) = compareRow; isUnusedSetting(compareRow) = 0; compareRow = frameSize; }
            
        } // end compare row loop
        
    } // end X row loop
    return(isUnusedSetting);
}


/*
 
 ----------------------------------------------------------------------------
                        frameArrange
 ----------------------------------------------------------------------------
 
 Description:
    This is an organization function. It takes the particle's current position,
    X, and arranges it in order of the frame arrangement.
 
 */
void qparticle::frameArrange(){
    
    MatrixXd Xscaled(rowLimit, nCol);
    Xscaled = X;
    for (int iRow = 0; iRow < rowLimit; iRow++){
        for (int iCol = nDis+1; iCol < nCol; iCol++){
            Xscaled(iRow, iCol) = 2*(Xscaled(iRow, iCol) - lb(iCol-1)) / ( ub(iCol-1) - lb(iCol-1) ) - 1;
        }
    }
    
    MatrixXi signs(rowLimit, nCol-1);
    for (int iRow = 0; iRow < rowLimit; iRow++){
        for (int iCol = 0; iCol < nCol-1; iCol++){
            signs(iRow, iCol) = 2 * (Xscaled(iRow, iCol+1) > 0) - 1;
        }
    }
    
    VectorXi belongsTo(rowLimit);
    
    int compareRow = 0;
    for (int iRow = 0; iRow < rowLimit; iRow++){
        
        // Get the sign of the first value; use it to pick a starting point
        if (signs(iRow, 0) > 0){ compareRow = frameSize / 2; } else {compareRow = 0;}
        
        for (compareRow; compareRow < frameSize; compareRow++){
            bool isSame = true;
            
            for (int iCol = 0; iCol < nDis; iCol++){
                
                if ( signs(iRow, iCol) != frameArrangement(compareRow, iCol) ){ iCol = nCol; isSame = false;}
            
            } // end column loop
            
            // If no differences were found, terminate search for this row;
            if (isSame){belongsTo(iRow) = compareRow; compareRow = frameSize;}
        
        } // end compare row loop
        
    } // end X row loop
    
    //******** Now rearrange ********
    MatrixXd tempX(rowLimit, nCol);
    VectorXd tempP(rowLimit);
    VectorXi sortedBelongsTo(rowLimit);
    
    sortedBelongsTo = belongsTo;
    std::sort(sortedBelongsTo.data(), sortedBelongsTo.data()+sortedBelongsTo.size());
    
    // Perform the row swapping
    for (int iRow = 0; iRow < rowLimit; iRow++){
        int rowRank = belongsTo(iRow);
        for (int switchRow = 0; switchRow < rowLimit; switchRow++){
            // Handle a match
            if (rowRank == sortedBelongsTo(switchRow)){
                sortedBelongsTo(switchRow) = -999;
                tempX.row(switchRow) = X.row(iRow);
                tempP(switchRow) = P(iRow);
                switchRow = rowLimit;
            }
        }
    }
    
    // Now arrange properly
    X = tempX;
    P = tempP;
    
    // Calculate the interactions
    MatrixXd interactionBlock(rowLimit, nBeta - nCol);
    for (int iRow = 0; iRow < rowLimit; iRow++){
        for (int iInt = 0; iInt < nInteraction; iInt++){
            double interactionValue = 1.0;
            for (int iCol = 1; iCol < nCol; iCol++){
                //cout << interactionMatrix(iInt, iCol) * X(iRow, iCol) << "\n";
                if (interactionMatrix(iInt, iCol) == 1){
                    interactionValue *= interactionMatrix(iInt, iCol) * X(iRow, iCol);
                }
            }
            interactionBlock(iRow, iInt) = interactionValue;
        }
    }
    fullX << X, interactionBlock;
    
}


/*
 
 ----------------------------------------------------------------------------
                        frameArrangeIndexSet
 ----------------------------------------------------------------------------
 
 Arguments:
    indexSet    -   Eigen vector object of support point indices to rearrange
 
 Description:
    This is an organization function. It takes a subset of the particle's current
    position, X, and arranges it in order of the frame arrangement.
 
 */

void qparticle::frameArrangeIndexSet(VectorXd indexSet){
    
    MatrixXd Xscaled(rowLimit, nCol);
    Xscaled = X;
    for (int iRow = 0; iRow < rowLimit; iRow++){
        for (int iCol = nDis+1; iCol < nCol; iCol++){
            Xscaled(iRow, iCol) = 2*(Xscaled(iRow, iCol) - lb(iCol-1)) / ( ub(iCol-1) - lb(iCol-1) ) - 1;
        }
    }
    
    MatrixXi signs(rowLimit, nCol-1);
    for (int iRow = 0; iRow < rowLimit; iRow++){
        for (int iCol = 0; iCol < nCol-1; iCol++){
            signs(iRow, iCol) = 2 * (Xscaled(iRow, iCol+1) > 0) - 1;
        }
    }
    
    VectorXi belongsTo(rowLimit);
    
    int compareRow = 0;
    for (int iRow = 0; iRow < rowLimit; iRow++){
        
        // Get the sign of the first value; use it to pick a starting point
        if (signs(iRow, 0) > 0){ compareRow = frameSize / 2; } else {compareRow = 0;}
        
        for (compareRow; compareRow < frameSize; compareRow++){
            bool isSame = true;
            
            for (int iCol = 0; iCol < nDis; iCol++){
                
                if ( signs(iRow, iCol) != frameArrangement(compareRow, iCol) ){ iCol = nCol; isSame = false;}
                
            } // end column loop
            
            // If no differences were found, terminate search for this row;
            if (isSame){belongsTo(iRow) = compareRow; compareRow = frameSize;}
            
        } // end compare row loop
        
    } // end X row loop
    
    //******** Now rearrange ********
    MatrixXd tempX(rowLimit, nCol);
    VectorXd tempP(rowLimit);
    VectorXi sortedBelongsTo(rowLimit);
    
    // For the factors which have just been redrawn, push them out
    int amtAdd = pow(2, nBeta);
    for (int iRow = 0; iRow < rowLimit; iRow++){
        if (indexSet(iRow) == 1){
            belongsTo(iRow) += amtAdd;
        }
    }
    
    sortedBelongsTo = belongsTo;
    std::sort(sortedBelongsTo.data(), sortedBelongsTo.data()+sortedBelongsTo.size());
    
    // Perform the row swapping
    for (int iRow = 0; iRow < rowLimit; iRow++){
        int rowRank = belongsTo(iRow);
        for (int switchRow = 0; switchRow < rowLimit; switchRow++){
            // Handle a match
            if (rowRank == sortedBelongsTo(switchRow)){
                sortedBelongsTo(switchRow) = -999;
                tempX.row(switchRow) = X.row(iRow);
                tempP(switchRow) = P(iRow);
                switchRow = rowLimit;
            }
        }
    }
    
    // Now arrange properly
    X = tempX;
    P = tempP;
    
    // Calculate the interactions
    MatrixXd interactionBlock(rowLimit, nBeta - nCol);
    for (int iRow = 0; iRow < rowLimit; iRow++){
        for (int iInt = 0; iInt < nInteraction; iInt++){
            double interactionValue = 1.0;
            for (int iCol = 1; iCol < nCol; iCol++){
                //cout << interactionMatrix(iInt, iCol) * X(iRow, iCol) << "\n";
                if (interactionMatrix(iInt, iCol) == 1){
                    interactionValue *= interactionMatrix(iInt, iCol) * X(iRow, iCol);
                }
            }
            interactionBlock(iRow, iInt) = interactionValue;
        }
    }
    fullX << X, interactionBlock;
    
}



/*
 
 ----------------------------------------------------------------------------
                            frameArragePbestX
 ----------------------------------------------------------------------------
 
 
 Description:
    This is an organization function. It performs the exact same function as
    frame arrange but operates on the pbest position instead of the current
    particle position.
 
 */
void qparticle::frameArragePbestX(){
    
    MatrixXd Xscaled(rowLimit, nCol);
    Xscaled = pbestX;
    for (int iRow = 0; iRow < rowLimit; iRow++){
        for (int iCol = nDis+1; iCol < nCol; iCol++){
            Xscaled(iRow, iCol) = 2*(Xscaled(iRow, iCol) - lb(iCol-1)) / ( ub(iCol-1) - lb(iCol-1) ) - 1;
        }
    }
    
    MatrixXi signs(rowLimit, nCol-1);
    for (int iRow = 0; iRow < rowLimit; iRow++){
        for (int iCol = 0; iCol < nCol-1; iCol++){
            signs(iRow, iCol) = 2 * (Xscaled(iRow, iCol+1) > 0) - 1;
        }
    }
    
    VectorXi belongsTo(rowLimit);
    
    int compareRow = 0;
    for (int iRow = 0; iRow < rowLimit; iRow++){
        
        // Get the sign of the first value; use it to pick a starting point
        if (signs(iRow, 0) > 0){ compareRow = frameSize / 2; } else {compareRow = 0;}
        
        for (compareRow; compareRow < frameSize; compareRow++){
            bool isSame = true;
            
            for (int iCol = 0; iCol < nCol-1; iCol++){
                
                if ( signs(iRow, iCol) != frameArrangement(compareRow, iCol) ){ iCol = nCol; isSame = false;}
                
            } // end column loop
            
            // If no differences were found, terminate search for this row;
            if (isSame){belongsTo(iRow) = compareRow; compareRow = frameSize;}
            
        } // end compare row loop
        
    } // end X row loop
    
    //******** Now rearrange ********
    MatrixXd tempX(rowLimit, nCol);
    VectorXd tempP(rowLimit);
    VectorXi sortedBelongsTo(rowLimit);
    
    sortedBelongsTo = belongsTo;
    std::sort(sortedBelongsTo.data(), sortedBelongsTo.data()+sortedBelongsTo.size());
    
    // Perform the row swapping
    for (int iRow = 0; iRow < rowLimit; iRow++){
        int rowRank = belongsTo(iRow);
        for (int switchRow = 0; switchRow < rowLimit; switchRow++){
            // Handle a match
            if (rowRank == sortedBelongsTo(switchRow)){
                sortedBelongsTo(switchRow) = -999;
                tempX.row(switchRow) = pbestX.row(iRow);
                tempP(switchRow) = pbestP(iRow);
                switchRow = rowLimit;
            }
        }
    }
    
    // Now arrange properly
    pbestX = tempX;
    pbestP = tempP;

}

/*
 
 ----------------------------------------------------------------------------
                            calcAttractor
 ----------------------------------------------------------------------------
 
 Arguments:
    gbestX  -   Eigen matrix object of the best design found by the swarm
    gbestP  -   Eigen vector object of the best proportions found by the swarm
 
 Description:
    This function calculates the local attractor for each particle. This is the
    position in the design space to which the particle is drawn.
 
 */

void qparticle::calcAttractor(MatrixXd gbestX, MatrixXd gbestP){
 
    std::random_device rdm;
    std::mt19937 generator(rdm());
    std::uniform_real_distribution<double> pdist(0,1);
    
    double rid = pdist(generator);
    double Rid = pdist(generator);
    double phi = ( 2.0*rid ) / (2.0*rid + 2.0*Rid);
    
    for (int iRow = 0; iRow < rowLimit; iRow++){
        
        // Local attractor for the discrete factors
        for (int iCol = 1; iCol < nDis+1; iCol++){

            // Now using uniform crossover
            double rid2 = pdist(generator);
            if (rid2 <= 0.5){ localAtt(iRow, iCol) = pbestX(iRow, iCol); }
            else { localAtt(iRow, iCol) = gbestX(iRow, iCol); }
            
        }
        
        // Local attractor for the continuous factors
        for (int iCol = 1+nDis; iCol < nCol; iCol++){
            rid = pdist(generator);
            Rid = pdist(generator);
            phi = ( 2.0*rid ) / (2.0*rid + 2.0*Rid);
            localAtt(iRow, iCol) = phi * pbestX(iRow, iCol) + (1.0 - phi)*(gbestX(iRow, iCol));
        }
        
        // Local attractor for the proportions IF continuous design
        if (allocation == "approx"){
            rid = pdist(generator);
            Rid = pdist(generator);
            phi = ( 2.0*rid ) / (2.0*rid + 2.0*Rid);
            localAttP(iRow) = phi * pbestP(iRow) + (1.0 - phi)*(gbestP(iRow));
        }
        
    } // End row loop
    
}

/*
 
 ----------------------------------------------------------------------------
                            updatePosition
 ----------------------------------------------------------------------------
 
 Arguments:
 mbestX  -  Eigen matrix object of the mean best design
 mbestP  -  Eigen vector object of the mean best proportion
 alpha   -  double object, the expansion-contraction coefficent
 
 Description:
    This function updates the position of the particle as a function
    of its local attractor and the mean best position.
 
 */

void qparticle::updatePosition(MatrixXd mbestX, VectorXd mbestP, double alpha){
    
    std::random_device rdm;
    std::mt19937 generator(rdm());
    std::uniform_real_distribution<double> pdist(0,1);
    
    // Incretment the limit on the prop allocated to each support point;
    if (plim < plimfinal){
        plim = plim + plimIncr;
    }
    if (plim > plimfinal){ plim = plimfinal;}
    // Decrease the minimum proportion that can be allocated to a support point
    if (pmin > 0){
        pmin = pmin - pminIncr;
    }
    if (pmin < 0){ pmin = 0.0; }
    
    double sign = -1.0;
    
    // vars for hamming distance discrete version
    double hDistance = 0;
    double b = 0;
    double pflip = 0;
    
    double uRand = 0; double randV = 0; double addToLocalAtt;
    
    randV = pdist(generator);
    
    for (int iRow = 0; iRow < rowLimit; iRow++){
        
        hDistance = 0;
        // Update for the discrete factors
        for (int iCol = 1; iCol < nDis+1; iCol++){
            
            // Performing the update using prob of change based on hamming distance
            hDistance = hDistance + ( abs(mbestX(iRow, iCol) - X(iRow, iCol)) ) / 2;
        }
        uRand = pdist(generator);
        b = alpha * hDistance * log(1.0 / uRand);
        pflip = b / nDis;
        if (pflip > 1.0){pflip = 1.0;}
        if (pflip < 0.0){pflip = 0.0;}
        //cout << "probability of a flip is:" << pflip << endl;
        // Now loop through and flip with probability pflip
        for (int iCol = 1; iCol < nDis+1; iCol++){
            uRand = pdist(generator);
            if (uRand < pflip){
                if (X(iRow, iCol) == 1){ X(iRow, iCol) = -1; }
                else { X(iRow, iCol) = 1; }
            }
        }
        

        // Update for the continuous factors
        for (int iCol = 1+nDis; iCol < nCol; iCol++){
            // Perform the draw
            if (randV >= 0.5){sign = -1.0;} else{sign = 1.0;}
            
            uRand = pdist(generator);
            addToLocalAtt = sign * alpha * abs( X(iRow, iCol) -  mbestX(iRow, iCol)) * log(1.0 / uRand);
            X(iRow, iCol) = localAtt(iRow, iCol) + addToLocalAtt;
        }
        
        // Update the proportions allocated
        if (allocation == "approx"){
            if (randV < 0.5){sign = -1.0;} else{sign = 1.0;}
            uRand = pdist(generator);
            addToLocalAtt = sign * alpha * abs( P(iRow) -  mbestP(iRow)) * log(1.0 / uRand);
            P(iRow) = localAttP(iRow) + addToLocalAtt;
        }
        
    } // End row loop
    
}

/*
 
 ----------------------------------------------------------------------------
                        checkFitnessPropOnly
 ----------------------------------------------------------------------------
 
 Arguments:
 mbestP  -  Eigen vector object of the mean best proportion
 alpha   -  double object, the expansion-contraction coefficent
 
 Description:
    This function checks the fitness of a particle's position using
    ONLY a new set of proportions instead of a new set of proportions
    and a new set of factor settings.
 
 */

void qparticle::checkFitnessPropOnly(VectorXd mbestP, double alpha){

    VectorXd xb(rowLimit);
    
    // Need to get the full design based on pbestX
    MatrixXd pbestXFull(rowLimit, nCol + nInteraction);
    MatrixXd interactionBlock(rowLimit, nBeta - nCol);
    for (int iRow = 0; iRow < rowLimit; iRow++){
        for (int iInt = 0; iInt < nInteraction; iInt++){
            double interactionValue = 1.0;
            for (int iCol = 1; iCol < nCol; iCol++){
                if (interactionMatrix(iInt, iCol) == 1){
                    interactionValue *= interactionMatrix(iInt, iCol) * pbestX(iRow, iCol);
                }
            }
            interactionBlock(iRow, iInt) = interactionValue;
        }
    }
    pbestXFull << pbestX, interactionBlock;
    
    xb = pbestXFull * beta;
    
    // Temporary W matrix for the fisher information calculation
    MatrixXd tempW (rowLimit, rowLimit);
    for (int iRow = 0; iRow < rowLimit; iRow++){
        for (int iCol = 0; iCol < rowLimit; iCol++){
            tempW(iRow, iCol) = 0;
        }
    }
    
    // Temporary Fisher Information Matrix
    MatrixXd tempFisher(nBeta, nBeta);
    
    // Calculate the weighted W matrix
    for (int ii = 0; ii < rowLimit; ii++){
        tempW(ii,ii) = exp(xb(ii)) / ( ( 1 + exp(xb(ii)) ) * ( 1+exp(xb(ii)) ) );
        tempW(ii, ii) = tempW(ii, ii) * P(ii);
    }
    
    //cout << "The pbest matrix we are working with is: \n" << pbestX;
    tempFisher = pbestXFull.transpose() * tempW * pbestXFull;
    
    double newFitness = 0;
    if (optCrit == "D"){
        newFitness = log( tempFisher.determinant() );
    }
    else if (optCrit == "A"){
        newFitness = -( tempFisher.inverse().trace() );
    }
    
    if (newFitness > pbest){
        //cout << "\n We got a new pbest by using only the updated proportions: " << newFitness << endl;
        pbest = newFitness;
        fitness = newFitness;
        pbestP = P;
    }
    
}

/*
 
 ----------------------------------------------------------------------------
                        checkPositionBounds
 ----------------------------------------------------------------------------
 
 Description:
    This function checks if any of the factor settings in X or proportions
    are out of the design space. If they are, the function brings them back to
    the boundary.
 
 */

void qparticle::checkPositionBounds(){
    
    std::random_device rdm;
    std::mt19937 generator(rdm());
    std::uniform_real_distribution<double> pdist(0,1);
    
    double sumP = 0.0;
    
    VectorXd oldp = P;
    
    for (int iRow = 0; iRow < rowLimit; iRow++){
        
        // fix for the continuous factors
        for (int iCol = 1+nDis; iCol < nCol; iCol++){
            if ( X(iRow, iCol) < lb(iCol-1) ){ X(iRow, iCol) = lb(iCol-1) + 0.0; }
            if ( X(iRow, iCol) > ub(iCol-1) ){ X(iRow, iCol) = ub(iCol-1) + 0.0;  }
        }
        
    } // End row loop for continuous factors
    
    // Check the proportions if the design is an appoximate design
    if (allocation == "approx"){
        for (int iRow = 0; iRow < rowLimit; iRow++){
            // changed this to work based on pmin 2/17
            if ( P(iRow) < pmin ){ P(iRow) = pmin; }
            if ( P(iRow) > plim ){ P(iRow) = plim;}
            sumP = sumP + P(iRow);
        }
        
        for (int iRow = 0; iRow < rowLimit; iRow++){
            P(iRow) = P(iRow) / sumP;
        }
        
        // Handle case where all P were negative -> NaN
        if ( P(0) != P(0) ){
            for (int iRow = 0; iRow < rowLimit; iRow++){
                P(iRow) = abs(oldp(iRow));
                sumP = sumP + P(iRow);
            }
            for (int iRow = 0; iRow < rowLimit; iRow++){
                P(iRow) = P(iRow) / sumP;
            }
        }
    }
    
    updateInteractions();
    
}

/*
 
 ----------------------------------------------------------------------------
                            redistributeProps
 ----------------------------------------------------------------------------
 
 Description:
    Test function to take any part of a proportion larger than 1/k, where
    k is the number of factors, and redistribute it to other support points.
    Not currently in use, but somewhat promising for increasing solution diversity.
 
 */

void qparticle::redistributeProps(){
    
    std::random_device rdm;
    std::mt19937 generator(rdm());
    std::uniform_real_distribution<double> pdist(0,1);
    
    //------------ code block to try to redistribute excess p
    // Added on 2/15 goal is to bring things down to plim to stop wasting iterations
    double epsilon=0.0; VectorXd isUnder(rowLimit);
    for (int iRow = 0; iRow < rowLimit; iRow++){
        if (P(iRow) < plim){ isUnder(iRow) = 1; }
        if (P(iRow) >= plim){
            epsilon = epsilon + ( P(iRow) - plim );
            P(iRow) = plim;
            isUnder(iRow) = 0;
        }
    }

    // while we have some leftover (epsilon), loop through possible design points and add some to them
    if (epsilon > 0){
        int cRow = 0; double available = 0; double draw = 0;
        double amt = 0; double leftover = epsilon;
        std::vector<int> goodIndex; goodIndex.resize( isUnder.sum() );
        int cindex = 0; double numCand = 0;
        for (int i = 0; i < rowLimit; i++){ if (isUnder(i) == 1){goodIndex[cindex] = i; cindex++; numCand = numCand + 1.0;}}
        while (epsilon > 0.0001){
            // draw from good Index
            cRow = goodIndex[ floor(pdist(generator) * (numCand) ) ];
            // find how much we can add
            available = plim - P( cRow );
            draw = pdist(generator);
            amt = draw * leftover;
            if (amt > epsilon){amt = epsilon;}
            
            if (amt > available){P(cRow) = plim; epsilon = epsilon - available;}
            else{
                P(cRow) = P(cRow) + amt;
                epsilon = epsilon - amt;
            }
            epsilon = 1.0 - P.sum();
            
        }
    }
    

}


/*
 
 ----------------------------------------------------------------------------
                        updateInteractions
 ----------------------------------------------------------------------------
 
 Description:
    This function takes the factor settings stored in X and uses them to
    calculate the complete design matrix that includes interactions, denoted by
    "fullX"
 
 */

void qparticle::updateInteractions(){
    
    // Calculate the interactions
    MatrixXd interactionBlock(rowLimit, nBeta - nCol);
    for (int iRow = 0; iRow < rowLimit; iRow++){
        for (int iInt = 0; iInt < nInteraction; iInt++){
            double interactionValue = 1.0;
            for (int iCol = 1; iCol < nCol; iCol++){
                if (interactionMatrix(iInt, iCol) == 1){
                    interactionValue *= interactionMatrix(iInt, iCol) * X(iRow, iCol);
                }
            }
            interactionBlock(iRow, iInt) = interactionValue;
        }
    }
    fullX << X, interactionBlock;
    
}

/*
 
 ----------------------------------------------------------------------------
                            updateFisher
 ----------------------------------------------------------------------------
 
 Description:
    This function updates the fisher information matrix based on the
    particle design, fullX. Note that this is only used for the local optimality
    approach.
 
 */

void qparticle::updateFisher(){
    
    VectorXd xb(rowLimit);
    xb = fullX * beta;
    
    // Calculate the weighted W matrix
    for (int ii = 0; ii < rowLimit; ii++){
        W(ii,ii) = exp(xb(ii)) / ( ( 1 + exp(xb(ii)) ) * ( 1+exp(xb(ii)) ) );
        W(ii, ii) = W(ii, ii) * P(ii);
    }
        
    fisher = fullX.transpose() * W * fullX;
    
};

/*
 
 ----------------------------------------------------------------------------
                            updateFitness
 ----------------------------------------------------------------------------
 
 Description:
    Function to update the particle's fitness for the local optimality
    approach
 
 */

void qparticle::updateFitness(){
    
    fitness = log(fisher.determinant());
    
};


/*
 
 ----------------------------------------------------------------------------
                            checkImprovement
 ----------------------------------------------------------------------------
 
 Description:
    This checks if a particle's current design is better than its personal best
    design by comparing the current fitness to pbest. If the current fitness is 
    better than the pbest position and fitness are updated to the current position
    and fitness.
 
 */

void qparticle::checkImprovement(){
    
    // Design is singular (p crept too high), restart particle
    if (fitness == 0){
        singularCounter = singularCounter + 1;
    } else{
        singularCounter = 0;
    }
    
    if (singularCounter == 5){
        REinitParticle();
    }
    
    if (fitness > pbest){
        pbest = fitness;
        
        if (allocation == "approx"){
            checkDuplicatePoints();
        }
        
        pbestX = X;
        pbestP = P;
    }
    
}

/*
 
 ----------------------------------------------------------------------------
                            verifyParticlePosition
 ----------------------------------------------------------------------------
 
 Description:
    Function to check if a particle is in an allowed position and then
    to order its support points. Not currently in use, but helpful for 
    debugging new functions.
 
 */

void qparticle::verifyParticlePosition(){
    
    checkPositionBounds();
    checkDuplicatePoints();
    frameArrange();
    
}

/*
 
 ----------------------------------------------------------------------------
                        checkDuplicatePoints
 ----------------------------------------------------------------------------
 
 Description:
    Function to check if any support points are repeated in the design. If they
    are repeated, then they are merged by combining the proportions allocated
    to each point.
 
 */

void qparticle::checkDuplicatePoints(){
    
    VectorXd needsReplacing(rowLimit);
    for (int iRow = 0; iRow < rowLimit; iRow++){ needsReplacing(iRow) = 0; }
    
    for (int comprow = 0; comprow < rowLimit; comprow++){
        for (int row = (comprow+1); row < rowLimit; row++){
            float rowcomp = 0;
            for (int col = 1; col < nCol; col++){
                rowcomp += abs(X(comprow,col) - X(row,col));
            }
            if (rowcomp < 0.05){
                // we need to figure out which had alarger prop allocated to it
                float diff = P(comprow) - P(row);
                int replace_index = row; //default to replacing the lower row
                int add_index = comprow;
                if (diff < 0){
                    replace_index = comprow;
                    add_index = row;
                }
                
                // add one row to the other 
                P(add_index) += P(replace_index);
                P(replace_index) = 0.0;
                
                // Update the vector of rows that need replacing
                needsReplacing(replace_index) = 1;
                
            } // end the fixing statement
        }
    }
    
    // Check if any of the design points should be replaced. If yes, create an array of indices
    VectorXi unusedFrames(frameSize);
    if (needsReplacing.sum() > 99){
        //cout << "is " << needsReplacing.sum();
        // create the replacement index array
        std::vector<int> needsReplace_arr; needsReplace_arr.resize( needsReplacing.sum() );
        int indexCount = -1;
        //cout << "The rows that need replacing are: " << endl;
        for (int iRow = 0; iRow < rowLimit; iRow++){
            if (needsReplacing(iRow) == 1){
                indexCount++;
                needsReplace_arr[indexCount] = iRow;
            }
        }
        // Now get the vector of unused frame combinations
        unusedFrames=getUnusedFrame();
        // Convert the vector into an array
        std::vector<int> availableFrames; availableFrames.resize( unusedFrames.sum() );
        indexCount = -1;
        for (int iRow = 0; iRow < frameSize; iRow++){
            if (unusedFrames(iRow) == 1){ indexCount++; availableFrames[indexCount] = iRow; }
        }
        // Run the function to regenerate design points
        reDrawDesignPoint( needsReplace_arr, availableFrames);
    }
    
    frameArrange();
    
}


/*
 
 ----------------------------------------------------------------------------
                        calc_Bayes_fitness
 ----------------------------------------------------------------------------
 
 Arguments:
    *IA - pointer to an integrationAbscissas object, used to perform an approximation
          the the expectation of the Fisher Information
 
 Description:
    This function calculates the Bayes fitness of a particle's design using the 
    approximation to the expected of the Fisher Information developed in Gotwalt (2009)
 
 */

void qparticle::calc_Bayes_fitness(integrationAbscissas *IA){
    
    double deter = 0;
    
    // Perform integration of the Bayesian D Criterion
    int nrad = 2;
    int nQ = 3;
    int nS = IA->getNS();
    
    MatrixXd tempInfo1(nBeta, nBeta);
    
    // Need to do the W calculation for this set of parameters
    VectorXd betaC1(nBeta);
    for (int iB=0; iB < nBeta; iB++){
        betaC1(iB) = IA->thetaSearch[0][0][0][iB];
    }
    MatrixXd W1 = MatrixXd::Zero(rowLimit, rowLimit);
    // Calculate the weighted W matrix
    for (int ii = 0; ii < rowLimit; ii++){
        double xb1 = fullX.row(ii) * betaC1;
        W1(ii,ii) = exp(xb1) / ( ( 1 + exp(xb1) ) * ( 1+exp(xb1) ) );
        W1(ii, ii) = W1(ii, ii) * P(ii);
    }
    
    // Find the determinant of the information matrix
    tempInfo1 = fullX.transpose() * W1 * fullX;
    double detI1 = log(tempInfo1.determinant() ) * IA->thetaSearchWeights[0][0];;
    deter = detI1;
    
    for (int r = 1; r < nrad; r++){
        for (int q = 0; q < nQ; q++){
            for (int s = 0; s < nS; s++){
                // variable for information matrix
                MatrixXd tempInfo(nBeta, nBeta);
                
                // Need to do the W calculation for this set of parameters
                VectorXd betaC(nBeta);
                for (int iB=0; iB < nBeta; iB++){
                    betaC(iB) = IA->thetaSearch[r][q][s][iB];
                }
                //Matrix<double, 4, 4> x;
                MatrixXd W11 = MatrixXd::Zero(rowLimit, rowLimit);
                for (int ii = 0; ii < rowLimit; ii++){
                    double xb1 = fullX.row(ii) * betaC;
                    W11(ii,ii) = exp(xb1) / ( ( 1 + exp(xb1) ) * ( 1+exp(xb1) ) );
                    W11(ii, ii) = W11(ii, ii) * P(ii);
                }
                
                // Find the determinant of the information matrix
                tempInfo = fullX.transpose() * W11 * fullX;
                double detI = tempInfo.determinant() ;
                deter = deter + IA->thetaSearchWeights[r][s] * log(detI);
                
            } // end abscissas loop
        }
    }
    
    fitness = deter;
    
}

/*
 
 ----------------------------------------------------------------------------
                        Printing Methods
 ----------------------------------------------------------------------------
 
 Description:
    The group of functions below are used to print different parts of the
    particle to the console.
 
 */

void qparticle::printX(){
    std::cout << "The design matrix for this particle is: \n" << X << endl;
}

void qparticle::printP(){
    std::cout << "The prop allocated for this particle is: \n" << P << endl;
}

void qparticle::printCurrentDesign(){
    cout << "The current design for this particle is: \n";
    for (int iRow = 0; iRow < rowLimit; iRow++){
        for (int iCol = 0; iCol < nCol; iCol++){
            cout << X(iRow, iCol) << " ";
        }
        cout << P(iRow) << endl;
    }
}

void qparticle::printAttractor(){
    cout << localAtt;
};

void qparticle::printQbits(){
    std::cout << "The qbits for this particle are: \n" << qbit << endl;
}

void qparticle::printPbest(){
    std::cout << "The pbest value for this particle is: \n" << pbest << endl;
};

void qparticle::printPbestP(){
    std::cout << "The pbest proportion allocation for this particle is: \n" << pbestP << endl;
};

void qparticle::printPbestX(){
    std::cout << "The pbest design for this particle is: \n" << pbestX << endl;
};

void qparticle::printPbestQbits(){
    std::cout << "The pbest qbits value for this particle is: \n" << pbestQbit << endl;
};

void qparticle::printFisher(){
    std::cout << "The fisher information matrix for this particle is: \n" << fisher << endl;
}

void qparticle::printBestDesign(){
    cout << "The best experimental design found by this particle is: \n";
    for (int iRow = 0; iRow < rowLimit; iRow++){
        if (pbestP(iRow) > 0){
            for (int iCol = 0; iCol < nCol; iCol++){
                cout << pbestX(iRow, iCol) << " ";
            }
            cout << pbestP(iRow) << endl;
        }
    }
    cout << "With a fitness value of: " << pbest << endl;
}

void qparticle::printSideBySideDesigns(){
    for (int iRow = 0; iRow < rowLimit; iRow++){
        // Print the Current
        for (int iCol = 0; iCol < nCol; iCol++){
            cout << X(iRow, iCol) << " ";
        }
        cout << P(iRow) << "     ";
        // Print the pbest
        for (int iCol = 0; iCol < nCol; iCol++){
            cout << pbestX(iRow, iCol) << " ";
        }
        cout << pbestP(iRow) << endl;
        
    }
    cout << "fitness: " << fitness << "          " << pbest << endl;
}

qparticle::~qparticle(){
    
}


/*
 
 ----------------------------------------------------------------------------
                            checkEquivalence
 ----------------------------------------------------------------------------
 
 Description: 
    Function to evaluate the sensitivity function for the particle's pbest
    design at the design points.
 
 
 */

// Function to check the equivalence theorem at the design points
bool qparticle::checkEquivalence(){
    
    // Return true if equivalence thm is satisfied at the design points
    bool satisfied = true;
    
    // Final value to be returned by the equivalence check. Is equal
    //  to w * X * I^-1 * X - p;
    double maxValue = -999;
    
    // Need the inverse of the Fisher information matrix
    
    VectorXd xb(rowLimit);
    
    // Need to get the full design based on pbestX
    MatrixXd pbestXFull(rowLimit, nCol + nInteraction);
    MatrixXd interactionBlock(rowLimit, nBeta - nCol);
    for (int iRow = 0; iRow < rowLimit; iRow++){
        for (int iInt = 0; iInt < nInteraction; iInt++){
            double interactionValue = 1.0;
            for (int iCol = 1; iCol < nCol; iCol++){
                if (interactionMatrix(iInt, iCol) == 1){
                    interactionValue *= interactionMatrix(iInt, iCol) * pbestX(iRow, iCol);
                }
            }
            interactionBlock(iRow, iInt) = interactionValue;
        }
    }
    pbestXFull << pbestX, interactionBlock;
    
    xb = pbestXFull * beta;
    
    // Calculate the weighted W matrix
    MatrixXd Wequiv(rowLimit, rowLimit);
    for (int r = 0; r < rowLimit; r++){
        for (int c =0; c < rowLimit; c++){
            Wequiv(r,c) = 0;
        }
    }
    
    for (int ii = 0; ii < rowLimit; ii++){
        Wequiv(ii,ii) = exp(xb(ii)) / ( ( 1 + exp(xb(ii)) ) * ( 1+exp(xb(ii)) ) );
        Wequiv(ii, ii) = Wequiv(ii, ii) * pbestP(ii);
    }
    
    MatrixXd bestFisher(nBeta, nBeta), inverseFisherInfo(nBeta, nBeta);
    bestFisher = pbestXFull.transpose() * Wequiv * pbestXFull;
    inverseFisherInfo = bestFisher.inverse();
    
    for (int iPoint = 0; iPoint < rowLimit; iPoint++){
        double newval = 0;
        if ( pbestP(iPoint) > 0.0001){
            double xprimeIx = pbestXFull.row(iPoint) * inverseFisherInfo * pbestXFull.row(iPoint).transpose();
            newval = (Wequiv(iPoint, iPoint) / pbestP(iPoint)) * xprimeIx;
            newval = newval - nBeta;
        }
        
        if (newval > maxValue ){
            maxValue = newval;
        }
        
        if (abs(newval) > 0.01){
            satisfied = false;
        }
        
    }
    
    return(satisfied);
}


/*
 
 ----------------------------------------------------------------------------
                        setPbestX
 ----------------------------------------------------------------------------
 
 Arguments:
    newX    - Eigen matrix object containing the new pbest design
 
 Description:
    Function to manually set the pbestX for a particle to a new design.
 
 
 */


void qparticle::setPbestX(MatrixXd newX){
    pbestX = newX;
}

/*
 
 ----------------------------------------------------------------------------
                        writeDesignToCSV
 ----------------------------------------------------------------------------
 
 Arguments:
    filepath    - Path to the folder where the design is to be saved
    filename    - filename for the aaved design
 
 Description:
    Function to save the best design found by the particle to a csv file.
 
 */

void qparticle::writeDesignToCSV(string filepath, string filename){
    
    string outputFile = filepath + filename + ".csv";
    
    ofstream myfile;
    myfile.open (outputFile);
    
    // Print the column names
    for (int iCol = 0; iCol < nCol; iCol++){
        myfile << "X"  << iCol << ",";
    }
    myfile << "p" << "\n";
    
    for (int iRow = 0; iRow < rowLimit; iRow++){
        if (pbestP(iRow) > 0.000001){
            for (int iCol = 0; iCol < nCol; iCol++){
                myfile << pbestX(iRow, iCol) << ",";
            }
            myfile << pbestP(iRow) << "\n";
        }
    }
    myfile.close();
    
};



