//
//  qparticle.hpp
//  QPSO


/*
 
 QPARTICLE
 
 -------------------------------------------------------------------------------------------
         Class managing an individual particle that searches for an optimal design
 -------------------------------------------------------------------------------------------
 
 Class Attributes:
    X                   - Eigen matrix object, design matrix
    fullX               - Eigen matrix object, design matrix WITH interactions included
    W                   - Matrix of g'^2
    qbit                - Eigen matrix storing bits. Used in generating starting discrete positions
    fisher              - Eigen matrix object containing the Fisher information matrix for the current design.
    pbestX              - Eigen matrix object containing the best design the particle has found
    pbestQbit           - Eigen matrix object containing the best qbit settings the particle has found
    localAtt            - Eigen matrix object containing the local attraction point for each factor
    interactionMatrix   - Eigen matrix object where each row is an interaction
    frameArrangement    - Eigen matrix object containing the frame arrangement (see fn documentation)
    P                   - Eigen vector object containing the current proportions for each support point
    pbestP              - Eigen vector object containing the best proportions for each support point
    beta                - Eigen vector object containing the nominal values for the parameters
    lb                  - Eigen vector object containing the lower bound for each factor
    ub                  - Eigen vector object containing the upper bound for each factor
    localAttP           - Eigen vector object containing the local attraction point for each proportion
    currentXRow         - Eigen vector object containing a single row of the design matrix
    nCont               - integer number of continuous factors in the model
    nDis                - integer number of discrete factors in the model
    rowLimit            - integer maximum number of support points
    nCol                - integer number of columns in the design matrix
    frameSize           - integer number of rows in the frame arrangement
    nInteraction        - integer number of interactions in the model
    singularCounter     - integer counter of the number of singular designs found by the particle
    maxit               - integer maximum number of iterations
    nBeta               - integer number of parameters in the model
    fitness             - double current fitness value of the particle's design
    pbest               - double best fitness found by the particle
    plim                - double maximum value a proportion allocated to a support point can take at iter
    plimfinal           - double maximum value a proportion allocated to a support point can take
    plimIncr            - double rate of increase for the proportion limit
    pmin                - double smallest value a proportion is allowed to take at this iteration
    pminIncr            - double rate of decrease for pmin
    allocation          - string "approx" or "exact" for design point allocation
    optCrit             - string "D" or "A" optimality. Currently only D is fully implemented
 
*/

#ifndef qparticle_hpp
#define qparticle_hpp

#include "qpso.h"

class integrationAbscissas;

using namespace Eigen;
using namespace std;

class qparticle{
    
    MatrixXd X, fullX, W, qbit, fisher, pbestX, pbestQbit, localAtt, interactionMatrix;
    MatrixXi frameArrangement;
    VectorXd P, pbestP, beta, lb, ub, localAttP;
    VectorXd currentXRow;
    int nCont, nDis, rowLimit, nCol, frameSize, nInteraction, singularCounter = 0, maxit = 0;
    int nBeta;
    double fitness= -21436587, pbest= -9089786756;
    double plim=0, plimfinal=0, pliminit=0, plimIncr = 0, pmin = 0, pminIncr = 0;
    std::string allocation;
    std::string optCrit;
    
public:
    qparticle();
    void initParticle(VectorXd lbd, VectorXd ubd, VectorXd betaVec, int num_discrete, int num_continuous, int num_interactions, MatrixXd interactions, int rowLimit, int maxiti, VectorXd lowerBoundBeta, VectorXd upperBoundBeta, std::string allocate, std::string optimalityCrit);
    void generateInitialLocation();
    void verifyParticlePosition();
    void REinitParticle();
    void reDrawProportions();
    void reDrawDesignPoint(std::vector<int> ind, std::vector<int> indSet);
    
    void calcAttractor(MatrixXd gbestX,  MatrixXd gbestP);
    void updatePosition(MatrixXd mbestX, VectorXd mbestP, double alpha);
    void checkPositionBounds();
    void redistributeProps();
    void updateInteractions();
    void updateFisher();
    void updateFitness();
    void checkFitnessPropOnly(VectorXd mbestP, double alpha);
    void checkImprovement();
    void checkDuplicatePoints();
    void generateFrameArrangement();
    void frameArrange();
    void frameArrangeIndexSet(VectorXd indexSet);
    VectorXi getUnusedFrame();
    void frameArragePbestX();
    void calc_Bayes_fitness(integrationAbscissas *IA);
    
    // Reset conditions
    bool examinePbestP();
    
    //      Setter Methods
    void setPbestX(MatrixXd newX);
    void setPbest(double newFitness){pbest = newFitness;};
    void setPbestP(VectorXd newP){pbestP = newP;}
    void setX( MatrixXd Xnew){X = Xnew;}
    void setP( VectorXd newP ){P = newP;}
    
    //      Printing Methods
    void printX();
    void printP();
    void printCurrentDesign();
    void printAttractor();
    void printQbits();
    void printPbest();
    void printPbestP();
    void printPbestX();
    void printPbestQbits();
    void printFisher();
    void printBestDesign();
    void printSideBySideDesigns();
    
    //      Getter Methods
    double getFitness(){return fitness;}
    double getPbest(){return pbest;}
    double getPlim(){return plim;}
    MatrixXd getLocalAttX(){return localAtt;}
    VectorXd getLocalAttP(){return localAttP;}
    MatrixXd getPbestX(){return pbestX;}
    VectorXd getPbestP(){return pbestP;}
    MatrixXd getPbestQbit(){return pbestQbit;}
    MatrixXd getX(){return X;}
    MatrixXd getQbit(){return qbit;}
    VectorXd getP(){return P;}
    bool checkEquivalence();
    
    // Saving the design
    void writeDesignToCSV(string filepath, string filename);
    
    ~qparticle();
};

#endif /* qparticle_hpp */
