//
//  qswarm.cpp
//  QPSO


#include "qswarm.hpp"
    
/*
 
 ----------------------------------------------------------------------------
                        initializeSwarm
 ----------------------------------------------------------------------------
 
 Arguments:
     numParticle                   - integer number of particles in the swarm
     maxiter                       - integer maximum number of iterations
     IAnew                         - integrationAbscissas object create for pseudo Bayesian optimality approach
     lbd                           - Eigen vector of lower bounds for each factor
     ubd                           - Eigen vector of upper bounds for each factor
     betaVec                       - Eigen vector of nominal parameter values
     num_discrete                  - integer number of discrete factors
     num_continuous                - integer number of continuous factors
     num_interactions              - integer number of interactions
     interactions                  - Eigen matrix of interactions
     rowLimiti                     - integer maximum number of support points
     eb_criti                      - integer how often to do elitist breeding
     jump_num                      - integer number of positions to breed (Rec is 1)
     optTypeI                      - optimality Type (local or pseudo Bayesian)
     lower_efficiency_bound        - double minimum value for the lower efficiency bound to terminate early
     lowerLimit                    - Eigen vector of lower limits for EW designs, no longer used
     upperLimit                    - Eigen vector of upper limits for EW designs, no longer used
     loggingSetting                - boolean. If true, pbests are stored from each iteration to file
     swarmNum                      - integer number to track which swarm this is
     allocate                      - string of "exact" or "approximate" design
     optimalityCrit                - string of D or A optimality. Currently only D is implemented

 Description:
    This function initializes a qswarm object, which is a quantum-behaved particle swarm.
 
 */

// Default class constructor
qswarm::qswarm(){
    
};

// Class deconstructor
qswarm::~qswarm(){
    
};

void qswarm::initilizeSwarm(int numParticle, int maxiter, integrationAbscissas* IAnew, VectorXd lbd, VectorXd ubd, VectorXd betaVec, int num_discrete, int num_continuous, int num_interactions, MatrixXd interactions, int rowLimiti, int eb_criti, int jump_num, std::string optTypeI, double lower_efficiency_bound, VectorXd lowerLimit, VectorXd upperLimit, bool loggingSetting, int swarmNum, std::string allocate, std::string optimalityCrit){
    
    // Use the input arguments to set class attributes
    nPart = numParticle;
    nCol = 1 + num_discrete + num_continuous;
    nBeta = (int)betaVec.size();
    nDis = num_discrete;
    maxit = maxiter;
    beta = betaVec;
    eb_crit = eb_criti;
    jumpNum = jump_num;
    optType = optTypeI;
    leb = lower_efficiency_bound;
    swarmNumber = swarmNum;
    
    // A or D optimality
    optCrit = optimalityCrit;
    
    nInt = num_interactions;
    interactionMatrix.resize(nInt, nCol);
    interactionMatrix = interactions;
    
    // Assign exact vs continuous allocation
    allocation = allocate;

    logging = loggingSetting;
    // Resize the logging information
    if (logging){
        gbest_log.resize(maxit+1);
        pbest_log.resize(maxit+1, numParticle);
        fitness_log.resize(maxit+1, numParticle);
    }
    
    llim.resize(nBeta); ulim.resize(nBeta);
    llim = lowerLimit;
    ulim = upperLimit;
    
    ub.resize(nCol-1); lb.resize(nCol-1);
    ub = ubd; lb = lbd;
    
    IA = IAnew;
    
    fracPart = 1.0 / numParticle;   // used to calculate mbest
    rowLimit = rowLimiti;
    
    // Resize the mbest and gbest objects
    mbestX.resize(rowLimit, nCol); gbestX.resize(rowLimit, nCol);
    mbestP.resize(rowLimit); gbestP.resize(rowLimit);
    
    // Resize the swarm size to the number of particles
    swarm = new qparticle *[numParticle];
    
    // Initialize each of the particles - local optimality
    if (optType == "local"){
        for (int part = 0; part < numParticle; part++){
            swarm[part] = new qparticle();
            swarm[part]->initParticle(lbd, ubd, betaVec, num_discrete, num_continuous, num_interactions, interactions, rowLimit, maxit, llim, ulim, allocation, optCrit);
            swarm[part]->updateInteractions();
            swarm[part]->updateFisher();
            //cout << swarm[part] -> getFitness() << endl;
            swarm[part]->updateFitness();
            //cout << swarm[part] -> getFitness() << endl;
            swarm[part]->checkImprovement();
            if (swarm[part]->getFitness() > gbest){
                gbest = swarm[part]->getFitness();
                gbestIndex = part;
            }
        }
    }
    
    
    // Initialize each of the particles - pseudo bayesian optimality
    if (optType == "bayes"){
        for (int part = 0; part < numParticle; part++){
            swarm[part] = new qparticle();
            swarm[part]->initParticle(lbd, ubd, betaVec, num_discrete, num_continuous, num_interactions, interactions, rowLimit, maxit, llim, ulim, allocation, optCrit);
            swarm[part]->updateInteractions();
            swarm[part]->calc_Bayes_fitness(IA);
            swarm[part]->checkImprovement();
            if (swarm[part]->getFitness() > gbest){
                gbest = swarm[part]->getFitness();
                gbestIndex = part;
            }
        }
    }
    
    // Get the global best design and p
    gbestX = swarm[gbestIndex]->getX();
    gbestP = swarm[gbestIndex]->getP();
    
    // Set the first value of all logging variables
    if (logging){
        gbest_log(0) = gbest;
        for (int iPart = 0; iPart < nPart; iPart++){
            pbest_log(iter, iPart) = swarm[iPart]->getPbest();
            fitness_log(iter, iPart) = swarm[iPart]->getFitness();
        }
    }
    
    
}


/*
 
 ----------------------------------------------------------------------------
                            updateMBest
 ----------------------------------------------------------------------------
 
 Description:
    This function calculates the mbest position (mainstream thought). It is calculated
    as the average of the pbest position for each particle. Positions corresponding to 
    discrete factors are rounded to the nearest factor setting.
 
 */


// Update the mbest position
void qswarm::updateMBest(){
    
    // Create a random uniform distribution to draw from
    std::random_device rdm;
    std::mt19937 generator(rdm());
    std::uniform_real_distribution<double> probability(0,1);
    
    iter = iter + 1;
    
    // Calculate mbest from the previous iteration
    mbestX = fracPart * swarm[0]->getPbestX();
    mbestP = fracPart * swarm[0]->getPbestP();
    for (int particle = 1; particle < nPart; particle++){
        mbestX = mbestX + fracPart * swarm[particle]->getPbestX();
        mbestP = mbestP + fracPart * swarm[particle]->getPbestP();
    }
    
    // if mbest at discrete < 0, then mbest = -1, else mbest = 1
    for (int iRow = 0; iRow < rowLimit; iRow++){
        for (int iCol = 1; iCol < 1+nDis; iCol++){
            if (mbestX(iRow, iCol) < 0){ mbestX(iRow, iCol) = -1; }
            else { mbestX(iRow, iCol) = 1; }
        }
    }
    
}

/*
 
 ----------------------------------------------------------------------------
                        jitterMBestP
 ----------------------------------------------------------------------------
 
 Description:
    Function to randomly alter the mbest elements corresponding to 
    proportions allocated to support points. The idea is to increase
    the variability of the solution. Currently not in use.
 
 */

void qswarm::jitterMBestP(){

    // Create a random uniform distribution to draw from
    std::random_device rdm;
    std::mt19937 generator(rdm());
    std::uniform_real_distribution<double> addProp(-0.05, 0.05);
    
    cout << "This is a jitter iteration \n";
    
    for (int iRow = 0; iRow < rowLimit; iRow++){
        mbestP(iRow) = mbestP(iRow) + addProp(generator);
        if (mbestP(iRow) < 0){mbestP(iRow) = -1.0 * mbestP(iRow);}
        else if (mbestP(iRow) > 1){mbestP(iRow) = 1;}
    }
    
}


/*
 
 ----------------------------------------------------------------------------
                    updateIteratorLocal
 ----------------------------------------------------------------------------
 
 Description:
    Performs a single-iteration update of every particle in the swarm for
    LOCAL optimality.
 
 */
void qswarm::updateIteratorLocal(){
    
    minval = 999;
    
    alpha = 1.2 - 0.8 * ( (iter - 1) / (maxit - 1) );
    
    // Loop over each particle and update
    for (int i=0;i<nPart;i++){
        
        swarm[i]->calcAttractor(gbestX, gbestP);
        swarm[i]->updatePosition(mbestX, mbestP, alpha);
        swarm[i]->checkPositionBounds();
        swarm[i]->updateFisher();
        swarm[i]->updateFitness();
        swarm[i]->checkImprovement();
        // Just look at the new proportions
        if (iter > 2 && allocation == "approx"){
          swarm[i]->checkFitnessPropOnly(mbestP, alpha);
        }
        
        if (swarm[i]->getFitness() > gbest){ gbestIndex = i; updateGbest = true; gbest = swarm[gbestIndex]->getFitness();}
        if (swarm[i]->getPbest() < minval){ minval = swarm[i]->getPbest(); worstIndex = i;}
        
        
    }
 
    // Check for improvement in the global best
    if (updateGbest){
        gbestX = swarm[gbestIndex]->getPbestX();
        gbestP = swarm[gbestIndex]->getPbestP();
        updateGbest = false;
    }


    
    // Perform elitist breeding
    if (iter % eb_crit == 0){
        elitist_breeding();
    }
    if (iter % eb_crit == 0 && allocation=="exact"){
        elitist_breeding_designPt();
    }
     
    // Update the log
    if (logging){
        gbest_log(0) = gbest;
        for (int iPart = 0; iPart < nPart; iPart++){
            pbest_log(iter, iPart) = swarm[iPart]->getPbest();
            fitness_log(iter, iPart) = swarm[iPart]->getFitness();
        }
    }
    
}


/*
 
 ----------------------------------------------------------------------------
                    updateIteratorPseudoBayes
 ----------------------------------------------------------------------------
 
 Description:
    Performs a single-iteration update of every particle in the swarm for
    PSEUDO Bayesian optimality.
 
 */

void qswarm::updateIteratorPseudoBayes(){
    
    // Initialize minval to a large value
    minval = 99999999999;
    
    alpha = 1.2 - 0.5 * ( (iter - 1.0) / (maxit - 1.0) );
    
    // Loop over and update each particle
    for (int i=0;i<nPart;i++){
        swarm[i]->calcAttractor(gbestX, gbestP);
        swarm[i]->updatePosition(mbestX, mbestP, alpha);
        swarm[i]->checkPositionBounds();
        swarm[i]->calc_Bayes_fitness(IA);
        swarm[i]->checkImprovement();
        
        if (swarm[i]->getFitness() > gbest){ gbestIndex = i; updateGbest = true; gbest = swarm[gbestIndex]->getFitness();}
        if (swarm[i]->getPbest() < minval){ minval = swarm[i]->getPbest(); worstIndex = i;}
        
    }
    
    // For for improvement in the global best
    if (updateGbest){
        gbestX = swarm[gbestIndex]->getPbestX();
        gbestP = swarm[gbestIndex]->getPbestP();
        updateGbest = false;
    }
    
    // Perform elitist breeding
    if (iter % eb_crit == 0){
        elitist_breeding();     }
    
    // Update the log
    if (logging){
        gbest_log(0) = gbest;
        for (int iPart = 0; iPart < nPart; iPart++){
            pbest_log(iter, iPart) = swarm[iPart]->getPbest();
            fitness_log(iter, iPart) = swarm[iPart]->getFitness();
        }
    }
    
}

/*
 
 ----------------------------------------------------------------------------
                            checkConvergence
 ----------------------------------------------------------------------------
 
 Description:
    Redraw the proportions at each support point if the swarm has already converged.
    This is not currently in use. Was intended to help better search the proportions.
 
 */

bool qswarm::checkConvergence(){
    
    if (gbest / minval > 0.97){
        for (int iPart =0; iPart < nPart; iPart++){
            if (iPart != gbestIndex){
                swarm[iPart]->reDrawProportions();
            }
        }
    }
    
    return(false);
}


/*
 
 ----------------------------------------------------------------------------
                        elitist_breeding
 ----------------------------------------------------------------------------
 
 Description:
    Perform the elitist breeding mutator. Replace randomly selected elements
    of particle's positions with the positions of other particles and check to
    see if the resulting design is an improvement over the previous design for
    that particle.
 
 */

void qswarm::elitist_breeding(){
    
    std::random_device rdm;
    std::mt19937 generator(rdm());
    std::uniform_real_distribution<double> pdist(0,1);
    
    // Initialize variables
    MatrixXd epool_i(rowLimit, nCol);
    VectorXd epoolP_i(rowLimit);
    MatrixXd epool_compare(rowLimit, nCol);
    VectorXd epoolP_compare(rowLimit);
    MatrixXd normedX (rowLimit, nCol);
    
    double tspn = 0; double newFitness = -999;
    
    bool EBoccured = false;
    bool propoccured = false;
    
    // Loop over each particle
    for (int iPart = 0; iPart < nPart; iPart++){
        
        // Reset the boolean variables
        EBoccured = false;
        propoccured = false;
        newFitness = -999;
        
        // Only perform switching if rand is smaller than jumping rate
        if ( pdist(generator) <= jumpPCT || (iPart == worstIndex && pdist(generator) <= 0.5) || (iPart == gbestIndex && pdist(generator) <= 0.5)){
            
            EBoccured = true;
            epoolP_i = swarm[iPart]->getPbestP();
            int iComp = ceil( (pdist(generator) * (nPart)) - 1 );
            tspn = pdist(generator);
            
            // Thinking about indices of a design X as
            //      1  -1  -1       0  1  2
            //      1  -1   1       3  4  5
            //      1   1  -1       6  7  8
            //      1   1   1       9 10 11
            
            // Get the edited index start point; create an index list, these act as starting pt
            std::vector< std::vector<int> > iTspn( jumpNum, std::vector<int> (2)  );
            int tspnIndexRow = floor( (pdist(generator) * (rowLimit)) );
            int tspnIndexCol = ceil( (pdist(generator) * (nCol-1)) );
            // Copy the index list; these will be incremented
            int iRowCurrent = tspnIndexRow; int iColCurrent = tspnIndexCol;
            // Loop and increment
            for (int iRow = 0; iRow < jumpNum; iRow++){
                iTspn[iRow][0] = iRowCurrent;
                iTspn[iRow][1] = iColCurrent;
                iColCurrent++;
                if (iColCurrent == nCol){
                    iColCurrent = 1;
                    iRowCurrent++;
                }
                if (iRowCurrent == rowLimit){
                    iRowCurrent = 0;
                }
            }
            
            // Those indices are to be moved somewhere else. Need to create that index list as well;
            std::vector< std::vector<int> > iTspnMove( jumpNum, std::vector<int> (2)  );
            int tspnIndexRowMove = floor( (pdist(generator) * (rowLimit)) );
            int tspnIndexColMove = ceil( (pdist(generator) * (nCol-1)) );
            // Copy the index list; these will be incremented
            int iRowCurrentMove = tspnIndexRowMove; int iColCurrentMove = tspnIndexColMove;
            // Loop and increment
            for (int iRow = 0; iRow < jumpNum; iRow++){
                iTspnMove[iRow][0] = iRowCurrentMove;
                iTspnMove[iRow][1] = iColCurrentMove;
                iColCurrentMove++;
                if (iColCurrentMove == nCol){
                    iColCurrentMove = 1;
                    iRowCurrentMove++;
                }
                if (iRowCurrentMove == rowLimit){
                    iRowCurrentMove = 0;
                }
            }
            
            // Finally need a list of indices to be shifted. This list will be all elements between
            // 1) the end of iTspnMove and the start of iTspn if iTspnMove "<" iTspn
            // 2) the end of iTspn and the start of iTspnMove if iTspn "<" iTspnMove
            int endOfTspnMove = nCol * iTspnMove[jumpNum-1][0] + iTspnMove[jumpNum-1][1];
            int startOfTspnMove = nCol * iTspnMove[0][0] + iTspnMove[0][0];
            int endOfTspn = nCol * iTspn[jumpNum-1][0] + iTspn[jumpNum-1][1];
            int startOfTspn = nCol * iTspn[0][0] + iTspn[0][0];
            bool overlap = false;
            if (startOfTspn > startOfTspnMove & startOfTspn < endOfTspnMove){ overlap = true; }
            if (startOfTspnMove > startOfTspn & startOfTspnMove < endOfTspn){ overlap = true; }
            
            // Handle operations within particle copy
            if (iComp == iPart){
                
                normedX = normalizeDesignEB( swarm[iPart]->getPbestX() );
                epool_i = normedX;
                epool_compare = normalizeDesignEB( swarm[iPart]->getPbestX() );
                
                // Cut and Paste - within
                if (tspn < 0){
                    
                } else {

                    for (int iRow = 0; iRow < jumpNum; iRow++){
                        epool_i(iTspnMove[iRow][0], iTspnMove[iRow][1]) = epool_compare( iTspn[iRow][0], iTspn[iRow][1]) ;
                    }

                    epool_i = revNormalizeDesignEB( epool_i );
                }
                
            } else {
                // Handle operations between particle copies
                
                normedX = normalizeDesignEB( swarm[iPart]->getPbestX() );
                epool_i = normedX;
                epool_compare = normalizeDesignEB( swarm[iComp]->getPbestX() );
                
                // Cut and Paste - Between particles
                if (tspn > 1){
                    
                    cout << "Warning: Something went wrong";
                    
                } else{
                    
                    //cout << "designw was \n" << swarm[iPart]->getPbestX();
                    for (int iRow = 0; iRow < jumpNum; iRow++){
                        epool_i(iTspnMove[iRow][0], iTspnMove[iRow][1]) = epool_compare( iTspn[iRow][0], iTspn[iRow][1]) ;
                    }
                
                    epool_i = revNormalizeDesignEB( epool_compare );

                }

            }
            
            
        } // end jump % check for the design poitns
        
        // Do another random draw, if satisfied perform elitist breeding on the proportions
        if (  (pdist(generator) <= jumpPCT || (iPart == worstIndex && pdist(generator) <= 0.5) || (iPart == gbestIndex && pdist(generator) <= 0.5)) && allocation == "approx" ){
            
            propoccured = true;
            VectorXd epoolP_i_compare(rowLimit);
            
            // If the first EB did not happen, then set epool_i to X
            if (!EBoccured){
                epool_i = swarm[iPart]->getPbestX();
            }
            
            EBoccured = true;
            epoolP_i = swarm[iPart]->getPbestP();
            int iComp = ceil( (pdist(generator) * (nPart)) - 1 );
            tspn = pdist(generator);
            
            // Get the edited index start point; create an index list, these act as starting pt
            std::vector<int> iTspn( jumpNum );
            int tspnIndexRow = floor( (pdist(generator) * (rowLimit)) );
            // Copy the index list; these will be incremented
            int iRowCurrent = tspnIndexRow;
            // Loop and increment
            for (int iRow = 0; iRow < jumpNum; iRow++){
                iTspn[iRow] = iRowCurrent;
                iRowCurrent++;
                if (iRowCurrent == rowLimit){
                    iRowCurrent = 0;
                }
            }
            
            // Those indices are to be moved somewhere else. Need to create that index list as well;
            std::vector<int> iTspnMove( jumpNum );
            int tspnIndexRowMove = floor( (pdist(generator) * (rowLimit)) );
            // Copy the index list; these will be incremented
            int iRowCurrentMove = tspnIndexRowMove;
            // Loop and increment
            for (int iRow = 0; iRow < jumpNum; iRow++){
                iTspnMove[iRow] = iRowCurrentMove;
                iRowCurrentMove++;
                if (iRowCurrentMove == rowLimit){
                    iRowCurrentMove = 0;
                }
            }
            
            // Finally need a list of indices to be shifted. This list will be all elements between
            // 1) the end of iTspnMove and the start of iTspn if iTspnMove "<" iTspn
            // 2) the end of iTspn and the start of iTspnMove if iTspn "<" iTspnMove
            int endOfTspnMove = iTspnMove[jumpNum-1];
            int startOfTspnMove = iTspnMove[0];
            int endOfTspn = iTspn[jumpNum-1];
            int startOfTspn = iTspn[0];
            bool overlap = false;
            if (startOfTspn > startOfTspnMove & startOfTspn < endOfTspnMove){ overlap = true; }
            if (startOfTspnMove > startOfTspn & startOfTspnMove < endOfTspn){ overlap = true; }
            
            // Perform the elitist breeding
            
            // Handle operations within particle copy
            if (iComp == iPart){
                
                epoolP_i = swarm[iPart]->getPbestP();
                epoolP_i_compare = swarm[iPart]->getPbestP();
                
                // Cut and Paste - within
                if (tspn < 0){
                    
                } else {
                    
                    for (int iRow = 0; iRow < jumpNum; iRow++){
                        epoolP_i(iTspnMove[iRow]) = epoolP_i_compare( iTspn[iRow]) ;
                    }
                    
                }
                
                
            } else {
                // Handle operations between particle copies
                
                epoolP_i = swarm[iPart]->getPbestP();
                epoolP_i_compare = swarm[iComp]->getPbestP();
                
                // Cut and Paste - Between particles
                if (tspn > 1){
                    
                    cout << "uh oh";
                    
                } else{
                    
                    //cout << "designw was \n" << swarm[iPart]->getPbestX();
                    for (int iRow = 0; iRow < jumpNum; iRow++){
                        epoolP_i(iTspnMove[iRow]) = epoolP_i_compare( iTspn[iRow]) ;
                    }
                    
                }
                
            }
            
            // Now need to constrain the proportions to sum to 1
            double sumP = epoolP_i.sum();
            for (int iRow = 0; iRow < rowLimit; iRow++){
                epoolP_i(iRow) = epoolP_i(iRow) / sumP;
            }

            
        } // end proportion elitist breeding
        
        // If breeding was performed, check the fitness of the new particle.
        if (EBoccured){
            
            // Check the fitness of the EB particle.
            if (optType == "local"){
                newFitness = check_epool_fitness(epool_i, epoolP_i);
            } else if (optType == "bayes"){
                newFitness = check_epool_fitness_pseudoBayes(epool_i, epoolP_i);
            }
            
            if (newFitness > swarm[iPart]->getPbest()){
                //cout << endl << "we have an elitist improvement for particle " << iPart << endl;
                swarm[iPart]->setX(epool_i);
                swarm[iPart]->setP(epoolP_i);
                if (allocation == "approx"){
                    swarm[iPart]->checkDuplicatePoints();
                }
                swarm[iPart]->frameArrange();
                swarm[iPart]->setPbestX( swarm[iPart]->getX() );
                swarm[iPart]->setPbestP( swarm[iPart]->getP() );
                swarm[iPart]->frameArragePbestX();
                swarm[iPart]->setPbest(newFitness);
                if (newFitness > gbest){
                    gbest = newFitness;
                    gbestX = swarm[iPart]->getPbestX();
                    gbestP = swarm[iPart]->getPbestP();
                    gbestIndex = iPart;
                    
                }
            }
        }
        
    } // end particle check
    
}

/*
 
 ----------------------------------------------------------------------------
                        redistributeProps
 ----------------------------------------------------------------------------
 
 Description:
    Function to redistribution proportions that go over 1/k where k is the
    number of factors. Not currently in use.
 
 */

VectorXd qswarm::redistributeProps( VectorXd eb_prop ){
    
    VectorXd P( eb_prop.size() );
    P = eb_prop;
    VectorXd isUnder( eb_prop.size() );
    
    // create the vector keeping track of eligable proportions
    double plim = 1.0 / nBeta;
    
    std::random_device rdm;
    std::mt19937 generator(rdm());
    std::uniform_real_distribution<double> pdist(0,1);
    
    //------------ code block to try to redistribute excess p
    double epsilon=0;
    for (int iRow = 0; iRow < rowLimit; iRow++){
        if (P(iRow) < plim){ isUnder(iRow) = 1; }
        if (P(iRow) >= plim){
            epsilon = epsilon + ( P(iRow) - plim );
            P(iRow) = plim;
            isUnder(iRow) = 0;
        }
    }
    // while we have some leftover (epsilon), loop through possible design points and add some to them
    if (epsilon > 0.0001){
        int cRow = 0; double available = 0; double draw = 0;
        double amt = 0; double leftover = epsilon;
        std::vector<int> goodIndex; goodIndex.resize( isUnder.sum() );
        int cindex = 0; double numCand = 0;
        for (int i = 0; i < rowLimit; i++){ if (isUnder(i) == 1){goodIndex[cindex] = i; cindex++; numCand = numCand + 1.0;}}
        while (epsilon > 0){
            // draw from good Index
            cRow = goodIndex[ floor(pdist(generator) * (numCand) ) ];
            // find how much we can add
            available = plim - P( cRow );
            draw = pdist(generator);
            amt = draw * leftover;
            if (amt > epsilon){amt = epsilon;}
            
            if (amt > available){P(cRow) = plim; epsilon = epsilon - available;}
            else{
                P(cRow) = P(cRow) + amt;
                epsilon = epsilon - amt;
            }
        }
    }
    
    return( P );
    
}

/*
 
 ----------------------------------------------------------------------------
                        check_epool_fitness
 ----------------------------------------------------------------------------
 
 Arguments:
    epool           - Eigen matrix object, is an experimental design
    eP              - Eigen vector object, the proporitons allocated to each support point
 
 Output:
    fitness         - double, value of the fitness criteria for the design epool, eP
 
 Description:
    Function to calculate the fitness criteria of the design presented as an argument
    to the function. The primary use of this function is for the elitist breeding operator
    (hence the eb and epool name), but the function can be used elsewhere.
 
 */

double qswarm::check_epool_fitness(MatrixXd epool, VectorXd eP){
    double fitness = 0;
    
    // Initialize the W matrix to be all zeros
    MatrixXd W (rowLimit, rowLimit);
    for (int iRow = 0; iRow < rowLimit; iRow++){
        for (int iCol = 0; iCol < rowLimit; iCol++){
            W(iRow, iCol) = 0.0;
        }
    }
    
    // Get the full design point including interaction
    
    MatrixXd fullX(rowLimit, nCol + nInt);
    // Calculate the interactions
    MatrixXd interactionBlock(rowLimit, nBeta - nCol);
    for (int iRow = 0; iRow < rowLimit; iRow++){
        for (int iInt = 0; iInt < nInt; iInt++){
            double interactionValue = 1.0;
            for (int iCol = 1; iCol < nCol; iCol++){
                //cout << interactionMatrix(iInt, iCol) * X(iRow, iCol) << "\n";
                if (interactionMatrix(iInt, iCol) == 1){
                    interactionValue *= interactionMatrix(iInt, iCol) * epool(iRow, iCol);
                }
            }
            interactionBlock(iRow, iInt) = interactionValue;
        }
    }
    fullX << epool, interactionBlock;
    
    // Calculate the diagonal elements of the W matrix
    for (int iRow = 0; iRow < rowLimit; iRow++){
        double xb = fullX.row(iRow) * beta;
        W(iRow, iRow) = eP(iRow) * exp(xb) / ( (1+exp(xb)) * (1+exp(xb)) );
    }
    
    // Calculate the fisher information matrix
    MatrixXd info (rowLimit, rowLimit);
    info = fullX.transpose() * W * fullX;
    
    fitness = log(info.determinant());
    
    return(fitness);
}

/*
 
 ----------------------------------------------------------------------------
                    check_epool_fitness_pseudoBayes
 ----------------------------------------------------------------------------
 
 Arguments:
 epool           - Eigen matrix object, is an experimental design
 eP              - Eigen vector object, the proporitons allocated to each support point
 
 Output:
 fitness         - double, value of the pseudo-Bayesian fitness criteria for the design epool, eP
 
 Description:
    Function to calculate the fitness criteria of the design presented as an argument
    to the function. The primary use of this function is for the elitist breeding operator
    (hence the eb and epool name), but the function can be used elsewhere. This function
    is the same as the previous function, but applies to the pseudo-Bayesian criteria
 
 */

double qswarm::check_epool_fitness_pseudoBayes(MatrixXd epool, VectorXd eP){
    double fitness = 0;
    
    double deter = 0;
    
    // Perform integration of the Bayesian D Criterion
    int nrad = 2;
    int nQ = 3;
    int nS = IA->getNS();
    
    // Perform sum for R = 0
    // variable for information matrix
    MatrixXd tempInfo1(nBeta, nBeta);
    
    // Need to do the W calculation for this set of parameters
    VectorXd betaC1(nBeta);
    for (int iB=0; iB < nBeta; iB++){
        betaC1(iB) = IA->thetaSearch[0][0][0][iB];
    }
    
    // Set up the full design with interaction
    MatrixXd fullX(rowLimit, nCol + nInt);
    // Calculate the interactions
    MatrixXd interactionBlock(rowLimit, nBeta - nCol);
    for (int iRow = 0; iRow < rowLimit; iRow++){
        for (int iInt = 0; iInt < nInt; iInt++){
            double interactionValue = 1.0;
            for (int iCol = 1; iCol < nCol; iCol++){
                if (interactionMatrix(iInt, iCol) == 1){
                    interactionValue *= interactionMatrix(iInt, iCol) * epool(iRow, iCol);
                }
            }
            interactionBlock(iRow, iInt) = interactionValue;
        }
    }
    fullX << epool, interactionBlock;
    
    MatrixXd W1 = MatrixXd::Zero(rowLimit, rowLimit);
    // Calculate the weighted W matrix
    for (int ii = 0; ii < rowLimit; ii++){
        double xb1 = fullX.row(ii) * betaC1;
        W1(ii,ii) = exp(xb1) / ( ( 1 + exp(xb1) ) * ( 1+exp(xb1) ) );
        W1(ii, ii) = W1(ii, ii) * eP(ii);
    }
    
    // Find the determinant of the information matrix
    tempInfo1 = fullX.transpose() * W1 * fullX;
    double detI1 = log(tempInfo1.determinant()) * IA->thetaSearchWeights[0][0];
    deter = detI1;
    
    for (int r = 1; r < nrad; r++){
        for (int q = 0; q < nQ; q++){
            for (int s = 0; s < nS; s++){
                // variable for information matrix
                MatrixXd tempInfo(nBeta, nBeta);
                
                // Need to do the W calculation for this set of parameters
                VectorXd betaC(nBeta);
                for (int iB=0; iB < nBeta; iB++){
                    betaC(iB) = IA->thetaSearch[r][q][s][iB];
                }
                MatrixXd W11 = MatrixXd::Zero(rowLimit, rowLimit);
                for (int ii = 0; ii < rowLimit; ii++){
                    double xb1 = fullX.row(ii) * betaC;
                    W11(ii,ii) = exp(xb1) / ( ( 1 + exp(xb1) ) * ( 1+exp(xb1) ) );
                    W11(ii, ii) = W11(ii, ii) * eP(ii);
                }
                
                // Find the determinant of the information matrix
                tempInfo = fullX.transpose() * W11 * fullX;
                double detI = tempInfo.determinant();
                deter = deter + IA->thetaSearchWeights[r][s] * log(detI);
                
            } // end abscissas loop
        }
    }
    
    fitness = deter;
    
    return(fitness);
}


/*
 
 ----------------------------------------------------------------------------
                            normalizeDesignEB
 ----------------------------------------------------------------------------
 
 Arguments:
    X                  - Eigen matrix object, is an experimental design
 
 Output:
    normDesign         - Eigen matrix object, normalized experimental design
 
 Description:
    The elitist breeding mutator pastes factor levels onto other factor levels
    the factor levels do not alwaays have the same range, thus the conversion is
    performed on the normalized (0,1) scale and then backtransformed to the true
    factor range using the revNormalizeDesignEB function.
 
*/

MatrixXd qswarm::normalizeDesignEB(MatrixXd X){
    MatrixXd normDesign (rowLimit, nCol);
    
    for (int iRow = 0; iRow < rowLimit; iRow++){
        normDesign(iRow, 0) = 1;
        for (int iCol = 1; iCol < nCol; iCol++){
            normDesign(iRow, iCol) = ( X(iRow, iCol) - lb(iCol-1) ) / ( ub(iCol - 1) - lb(iCol - 1) );
        }
    }
    
    return(normDesign);
}

/*
 
 ----------------------------------------------------------------------------
                      revNormalizeDesignEB
 ----------------------------------------------------------------------------
 
 Arguments:
    X                  - Eigen matrix object, normalized experimental design
 
 Output:
    revNormDesign      - Eigen matrix object, experimental design in original scale
 
 Description:
    Function that takes a normalized experimental design (see normalizeDesignEB function)
    and converts it back to the original factor ranges.
 
*/

MatrixXd qswarm::revNormalizeDesignEB(MatrixXd X){
    MatrixXd revNormDesign (rowLimit, nCol);
    
    // Reverse the normalization
    for (int iRow = 0; iRow < rowLimit; iRow++){
        revNormDesign(iRow, 0) = 1.0;
        // Correct the discrete factors
        for (int iCol = 1; iCol < nDis+1; iCol++){
            if (X(iRow, iCol) < 0.5){
                revNormDesign(iRow, iCol) = lb(iCol-1);
            } else{
                revNormDesign(iRow, iCol) = ub(iCol-1);
            }
        }
        // Correct the continuous factors
        for (int iCol = nDis+1; iCol < nCol; iCol++){
            revNormDesign(iRow, iCol) = X(iRow, iCol)*(ub(iCol-1)-lb(iCol-1)) + lb(iCol-1) ;
        }
    }
    
    return(revNormDesign);
}

/*
 
 ----------------------------------------------------------------------------
                            tweakProps
 ----------------------------------------------------------------------------
 
 Description:
    Experimental mutator function. Takes the set of proportions and fuzzes them
    slightly to get a new set that is similar to the original set. Not currently
    in use.
 
*/

void qswarm::tweakProps(){
    
    // Set up random number generation
    std::random_device rdm;
    std::mt19937 generator(rdm());
    std::uniform_real_distribution<double> pdist(-0.05,0.05);
    std::uniform_real_distribution<double> indexDist(0.0,1.0);
    
    VectorXd tempP(rowLimit);
    tempP = swarm[gbestIndex]->getPbestP();
    
    // Starting with just a fixed increase
    int pIndex = floor( (indexDist(generator) * (rowLimit)) );
    double newVal = pdist(generator);
    
    // Add the drawn value to the selected index
    tempP(pIndex) += newVal;
    if (tempP(pIndex) < 0){tempP(pIndex) = 0;}
    //cout << "TempP is: " << tempP(pIndex) << endl;
    
    // Adjust all remaining proportions
    double sumP = 0.0;
    double leftover = 1.0 - tempP(pIndex);
    for (int iRow = 0; iRow < rowLimit; iRow++){
        if (iRow != pIndex){
            sumP += tempP(iRow);
        }
    }
    
    for (int iRow = 0; iRow < rowLimit; iRow++){
        if (iRow != pIndex){
            tempP(iRow) = (tempP(iRow) / sumP) * leftover;
        }
    }
    
    // Now evaluate the fitness at the adjusted proportion
    double newFitness = 0;
    newFitness = check_epool_fitness(swarm[gbestIndex]->getPbestX(), tempP);
    
    // Check for improvement
    if (newFitness > gbest){
        gbestP = tempP;
        swarm[gbestIndex]->setPbestP(tempP);
        swarm[gbestIndex]->setPbest(newFitness);
        gbest = newFitness;
    }
    
};

/*
 
 ----------------------------------------------------------------------------
                            random_rounding
 ----------------------------------------------------------------------------
 
 Description:
    Experimental mutator function. Takes the set of proportions and rounds them
    to randomly selected decimal places (place chosen using a poisson dist draw).
    Goal is to find new candidate values for the proportion set. Not currently in
    use.
 
 */

void qswarm::random_rounding(){
    
    // Set up the two poisson distributions
    std::random_device rdm;
    std::mt19937 generator(rdm());
    std::poisson_distribution<int> poi1dist(1);
    std::uniform_int_distribution<int> unif(0,rowLimit-1);
    std::poisson_distribution<int> poi3dist(3);
    
    int nRound = 0; // number of proportions to be rounded.
    int placeRound = 0; // place to be rounded to
    int changeIndex = 0; // index to be rounded
    VectorXd newProps(rowLimit); // copy of the proportions to this
    double newFitness = 0;
    
    // Loop over each particle
    for (int iPart = 0; iPart < nPart; iPart++){
        
        // Copy over the pbest proportions
        newProps = swarm[iPart]->getPbestP();
        
        // Might add a random draw here to determine if rounding occurs
        
        // Draw the number of proportions to round and the place to round them to.
        nRound = poi1dist(generator);
        placeRound = poi3dist(generator);
        
        // For each nRound, draw a design point and round it there. Note that it is possible to round
        //  the same point more than one time.
        for (int iPoint = 0; iPoint < nRound; iPoint++){
            
            // get the index
            changeIndex = unif(generator);
            newProps(changeIndex) = round( (newProps(changeIndex) * pow(10, placeRound) ) ) / pow(10, placeRound);
            
        }
        
        // Now need to constrain the proportions to sum to 1
        double sumP = newProps.sum();
        for (int iRow = 0; iRow < rowLimit; iRow++){
            newProps(iRow) = newProps(iRow) / sumP;
        }
        
        //cout << "\n The new props are: " << newProps;
        
        // Now we want to test for improvement, using the old epool function
        if (optType == "local"){
            newFitness = check_epool_fitness( swarm[iPart]->getPbestX(), newProps);
        } else if (optType == "bayes"){
            newFitness = check_epool_fitness_pseudoBayes( swarm[iPart]->getPbestX(), newProps );
        }
        
        //cout << "the fitness for this si: " << newFitness << endl;
        
        if (newFitness > swarm[iPart]->getPbest() ){
            swarm[iPart]->setP(newProps);
            if (allocation == "approx"){
                swarm[iPart]->checkDuplicatePoints();
            }
            swarm[iPart]->frameArrange();
            swarm[iPart]->setPbestX( swarm[iPart]->getPbestX() );
            swarm[iPart]->setPbestP( swarm[iPart]->getP() );
            swarm[iPart]->frameArragePbestX();
            swarm[iPart]->setPbest(newFitness);
            
            if (newFitness > gbest){
                gbest = newFitness;
                gbestX = swarm[iPart]->getPbestX();
                gbestP = swarm[iPart]->getPbestP();
                gbestIndex = iPart;
            }
            
        }
        
    } // end loop over each particle
    
}


/*
 
 ----------------------------------------------------------------------------
                        elitist_breeding_designPt
 ----------------------------------------------------------------------------
 
 Description:
    Variant of the elitist breeding mutator. Instead of changing only a single
    position, the entire set of discrete factor settings for the selected
    design point is switched with another set. Not currently in use.
 
 */

void qswarm::elitist_breeding_designPt(){
    
    std::random_device rdm;
    std::mt19937 generator(rdm());
    std::uniform_real_distribution<double> pdist(0,1);
    
    int iComp = 0;
    int iRowCopy = 0;
    int iRowToOverwrite = 0;
    
    double newFitness = 0;
    
    // Empty matrix to store the temporary design design
    MatrixXd tempX(rowLimit, nCol);
    VectorXd tempP(rowLimit);
    for (int iRow = 0; iRow < rowLimit; iRow++){ tempP(iRow) = 1; }
    
    for (int iPart = 0; iPart < nPart; iPart++){
        
        // Assign tempX to the pbest position
        tempX = swarm[iPart]->getPbestX();
        
        // Select a particle to breed with at random
        iComp = ceil( (pdist(generator) * (nPart)) - 1 );
        // Select a row at random to be overwritten
        iRowToOverwrite = floor( pdist(generator) * (rowLimit) );
        // Select a row at random to copy over
        iRowCopy = floor( pdist(generator) * (rowLimit) );
        
        tempX.row(iRowCopy) = swarm[iComp]->getPbestX().row(iRowCopy);
        
        newFitness = check_epool_fitness(tempX, tempP);
        
        if ( newFitness > swarm[iPart]->getPbest() ){
            
            swarm[iPart]->setX(tempX);
            swarm[iPart]->frameArrange();
            swarm[iPart]->setPbestX( swarm[iPart]->getX() );
            swarm[iPart]->frameArragePbestX();
            swarm[iPart]->setPbest(newFitness);
            if (newFitness > gbest){
                gbest = newFitness;
                gbestX = swarm[iPart]->getPbestX();
                gbestIndex = iPart;
            }
            
            
        }
        
    }
    
}

/*
 
 ----------------------------------------------------------------------------
                        getGbestFisherInformation
 ----------------------------------------------------------------------------
 
 Output:
    info     - Eigen matrix object, Fisher information matrix for the gbest design
 
 Description:
    Function to return the Fisher information matrix for the global
    best found design.
 
 */

MatrixXd qswarm::getGbestFisherInformation(){
    
    // Initialize the W matrix to be all zeros
    MatrixXd W (rowLimit, rowLimit);
    for (int iRow = 0; iRow < rowLimit; iRow++){
        for (int iCol = 0; iCol < rowLimit; iCol++){
            W(iRow, iCol) = 0.0;
        }
    }
    
    // Calculate the full design based on gbestX
    MatrixXd gbestXFull(rowLimit, nCol + nInt);
    MatrixXd interactionBlock(rowLimit, nBeta - nCol);
    for (int iRow = 0; iRow < rowLimit; iRow++){
        for (int iInt = 0; iInt < nInt; iInt++){
            double interactionValue = 1.0;
            for (int iCol = 1; iCol < nCol; iCol++){
                if (interactionMatrix(iInt, iCol) == 1){
                    interactionValue *= interactionMatrix(iInt, iCol) * gbestX(iRow, iCol);
                }
            }
            interactionBlock(iRow, iInt) = interactionValue;
        }
    }
    gbestXFull << gbestX, interactionBlock;
    
    // Calculate the diagonal elements of the W matrix
    for (int iRow = 0; iRow < rowLimit; iRow++){
        double xb = gbestXFull.row(iRow) * beta;
        W(iRow, iRow) = gbestP(iRow) * exp(xb) / ( (1+exp(xb)) * (1+exp(xb)) );
    }
    
    // Calculate the fisher information matrix
    MatrixXd info (nBeta, nBeta);
    info = gbestXFull.transpose() * W * gbestXFull;
    
    return(info);
    
}


/*
 
 ----------------------------------------------------------------------------
                                setGbest
 ----------------------------------------------------------------------------
 
 Output:
    particleNumber     - integer id of the particle to manually set the gbest from
    newFitness         - double value of the fitness of the new design
 
 Description:
    Function to manually set the swarm gbest to the design obtained by the specified
    particle.
 
 */

void qswarm::setGbest(int particleNumber, double newFitness){
    gbestIndex = particleNumber;
    gbest = newFitness;
    gbestX = swarm[particleNumber]->getPbestX();
    gbestP = swarm[particleNumber]->getPbestP();
};


/*
 
 ----------------------------------------------------------------------------
                            displayWorstInformation
 ----------------------------------------------------------------------------
 
 Description:
    Function to print to console the informatino about the worst particle in the swarm
 
 */

// Prints out a load of information about the worst particle in the swarm. Trying to figure out
//  why it seems to settle on one particle that then does not improve much.
void qswarm::displayWorstInformation(){
    cout << "\n ---------------------------------------- \n";
    cout << "It is iteration " << iter << "plim" << swarm[worstIndex]->getPlim() << " and the worst particle is " << worstIndex << endl;
    cout << "The current position of this particle is: \n" << swarm[worstIndex]->getX() << endl;
    cout << "The proportions allocated to the current position are: \n" << swarm[worstIndex]->getP() << endl;
    cout << "The fitness associated with this position is: " << swarm[worstIndex]->getFitness() << endl;
    cout << "The local attractor for this position was: \n" << swarm[worstIndex]->getLocalAttX() << endl;
    cout << "The prop local attractor for this position was: \n" << swarm[worstIndex]->getLocalAttP() << endl;
    cout << "The best position this particle has achieved is: \n" << swarm[worstIndex]->getPbestX() << endl;
    cout << "The proportions allocated to the best position are: \n" << swarm[worstIndex]->getPbestP() << endl;
    cout << "The fitness associated with the pbest position is: " << swarm[worstIndex]->getPbest();
    cout << "Compare this to the global best:" << endl;
    cout << "The gbest position is: \n" << swarm[gbestIndex]->getPbestX() << endl;
    cout << "The proportions allocated to the current position are: \n" << swarm[gbestIndex]->getPbestP() << endl;
    cout << "The fitness associated with this position is: " << gbest << endl;
    cout << "\n ---------------------------------------- \n";
}


/*
 
 ----------------------------------------------------------------------------
                            printGbestDesign
 ----------------------------------------------------------------------------
 
 Description:
    Function to print to console the best design found by the swarm.
 
 */

void qswarm::printGbestDesign(){
    
    int numDesignPoints = 0;
    double minimumP = 1;
    
    cout << "The best experimental design found by this swarm is: \n";
    for (int iRow = 0; iRow < rowLimit; iRow++){
        if (gbestP(iRow) > -0.5){
            numDesignPoints++;
            if (gbestP(iRow) < minimumP){ minimumP = gbestP(iRow);}
            for (int iCol = 0; iCol < nCol; iCol++){
                cout << gbestX(iRow, iCol) << " ";
            }
            cout << gbestP(iRow) << endl;
        }
    }
    cout << "With a fitness value of: " << gbest << endl;
    cout << "Design Points: " << numDesignPoints << "; minimum proportion: " << minimumP << endl;
}


/*
 
 ----------------------------------------------------------------------------
                         saveGbestDesign
 ----------------------------------------------------------------------------
 
 Arguments:
    filepath    - string path to folder where the design should be saved
    prefix      - string prefix for the filename
 
 Description:
    Function to save the best obtained design to a csv file.
 
 */

void qswarm::saveGbestDesign(string filepath, string prefix){
    string outputFileName = prefix + "_gbestDesign";
    swarm[gbestIndex]->writeDesignToCSV(filepath, outputFileName);
};

/*
 
 ----------------------------------------------------------------------------
                            saveGbestLog
 ----------------------------------------------------------------------------
 
 Arguments:
    filepath    - string path to folder where the file should be saved
    filename    - string filename for the log
 
 Description:
    Function to save the logged global best information to a csv file.
 
*/

void qswarm::saveGbestLog(string filepath, string filename){
    
    string outputFile = filepath + filename + "_gbestLog.csv";
    
    ofstream myfile;
    myfile.open (outputFile);
    
    for (int iIter = 0; iIter < maxit; iIter++){
        myfile << gbest_log(iIter) << ",";
    }
    
    myfile.close();
    
};

/*
 
 ----------------------------------------------------------------------------
                        writePbestLogToCSV
 ----------------------------------------------------------------------------
 
 Arguments:
    filepath    - string path to folder where the file should be saved
    filename    - string filename for the log
 
 Description:
    Function to save the logged pbest information to a csv file. Note
    that for large number of particles or large number of iterations
    this file can get quite large.
 
 */

void qswarm::writePbestLogToCSV(string filepath, string filename){
    
    string outputFile = filepath + filename + "_pbestLog.csv";
    
    ofstream myfile;
    myfile.open (outputFile);
    
    // Print the column names
    for (int iRow = 0; iRow < maxit; iRow++){
        for (int iCol = 0; iCol < nPart-1; iCol++){
            myfile << pbest_log(iRow, iCol) << ",";
        }
        myfile << pbest_log(iRow, nPart-1) << "\n";
    }
    myfile.close();
    
};

/*
 
 ----------------------------------------------------------------------------
                        writeFitnessLogToCSV
 ----------------------------------------------------------------------------
 
Arguments:
    filepath    - string path to folder where the file should be saved
    filename    - string filename for the log
 
 Description:
    Writes the fitness for each particle at each iteration to a csv file.
 
 */

void qswarm::writeFitnessLogToCSV(string filepath, string filename){
    
    string outputFile = filepath + filename + "_fitnessLog.csv";
    
    ofstream myfile;
    myfile.open (outputFile);
    
    // Print the column names
    
    for (int iRow = 0; iRow < maxit; iRow++){
        for (int iCol = 0; iCol < nPart-1; iCol++){
            myfile << fitness_log(iRow, iCol) << ",";
        }
        myfile << fitness_log(iRow, nPart-1) << "\n";
    }
    myfile.close();
    
};



