//
//  qswarm.hpp
//  QPSO

/*
    QSWARM
 
 -------------------------------------------------------------------------------------------
 Class will contain an object called swarm. Each element of the swarm is a qparticle object.
 -------------------------------------------------------------------------------------------
 
 Class Attributes:
    **swarm     :    An array of pointers to qparticle objects
    mbestX      :    The mean best position. Used to update the qparticle position
    mbestP      :    The mean best proportion allocated to each support point.
    gbestX      :    The overall best position (design points)
    gbestP      :    The overall best proportion allocated to each support point.
    pbest_log   :    Keeps track of each particle's pbest fitness across every iteration.
                     Only used if logging is turned on
    fitness_log :    Keeps track of each particle's fitness across every iteration. Only used if logging is
                     turned on
    gbest_log   :    Keeps track of the global best fitness across every iteration. Only used if logging is
                        turned on
    beta        :    Vector of the nominal beta values. Only needed for local optimality.
    llim        :    The lower limit for each parameter. Only used in EW designs.
    ulim        :    The upper limit for each parameter. Only used in EW designs.
    lb          :    The lower limit for each design point. Note that this does not include the intercept.
    ub          :    The upper limit for each design point. Note that this does not include the intercept.
 
 */

#ifndef qswarm_hpp
#define qswarm_hpp

#include "qpso.h"

class qswarm{

public:
    
    // Class member declaration
    qparticle **swarm;
    
    MatrixXd mbestX; MatrixXd gbestX; MatrixXd pbest_log; MatrixXd fitness_log;
    VectorXd mbestP, gbestP, beta, llim, ulim, gbest_log; VectorXd lb; VectorXd ub;
    MatrixXd interactionMatrix;
    
    bool logging; bool updateGbest = false;
    
    std::string allocation;
    std::string optCrit;
    
    int nPart, nCol, nBeta; int gbestIndex; int worstIndex; int oldWorstIndex = 0; int eb_crit;
    int nInt; int swarmNumber;
    int worstCounter = 0;
    int iteration = 0; int maxit; int rowLimit; int nDis; int iter = 0;
    
    double gbest = -12345678; double minval = 999; double fracPart; double alpha = 1; double leb;
    double jump_rate = 0.5; double jumpPCT = 0.25; int jumpNum = 1;
    
    integrationAbscissas* IA;

    std::string optType;
    
    // Class Constructor and initializer
    qswarm();
    void initilizeSwarm(int numParticle, int maxit, integrationAbscissas* IAnew, VectorXd lbd, VectorXd ubd, VectorXd betaVec, int num_discrete, int num_continuous, int num_interaction, MatrixXd interactions, int rowLimit, int eb_criti, int jump_num, std::string optTypeI, double lower_efficiency_bound, VectorXd lowerLimit, VectorXd upperLimit, bool loggingSetting, int swarmNum, std::string allocate, std::string optimalityCrit);
    
    // Updating methods
    void updateMBest();
    void jitterMBestP();
    void updateIteratorPseudoBayes();
    void updateIteratorLocal();
    bool checkConvergence();
    double check_epool_fitness(MatrixXd epool, VectorXd eP);
    double check_epool_fitness_pseudoBayes(MatrixXd epool, VectorXd eP);
    bool checkEquivalence(){ return(swarm[gbestIndex]->checkEquivalence()); }
    
    // Mutators
    void elitist_breeding();
    void random_rounding();
    MatrixXd normalizeDesignEB(MatrixXd X);
    MatrixXd revNormalizeDesignEB(MatrixXd X);
    void tweakProps();
    VectorXd redistributeProps( VectorXd eb_prop );
    
    void elitist_breeding_designPt();
    
    // Setter methods
    void setGbest(int particleNumber, double newFitness);
    void setGbestDirect( double newFitness ){ gbest = newFitness; };
    void setGbestX( MatrixXd newX ){ gbestX = newX; }
    void setGbestP( VectorXd newP ){ gbestP = newP; }
    void setGbestIndex( int newIndex ){ gbestIndex = newIndex; }
    
    // Getter methods
    MatrixXd getGbestFisherInformation();
    MatrixXd getGbestX(){ return gbestX; };
    VectorXd getGbestP(){ return gbestP; };
    double getGbest(){ return gbest; }
    int getGBestIndex(){ return gbestIndex; }
    VectorXd getPbestP(int iPart){return swarm[iPart]->getPbestP();}
    MatrixXd getPbestX(int iPart){return swarm[iPart]->getPbestX();}
    VectorXd getP(int iPart){return swarm[iPart]->getP();}
    MatrixXd getX(int iPart){return swarm[iPart]->getX();}
    double getPbest(int iPart){return swarm[iPart]->getPbest();}
    
    // Setter Methods for individual particles
    void setPbestP(int iPart, VectorXd tempP){swarm[iPart]->setPbestP(tempP);}
    void setPbestX(int iPart, MatrixXd tempX){swarm[iPart]->setPbestX(tempX);}
    void setPbest(int iPart, double tempFitness){swarm[iPart]->setPbest(tempFitness);}
    void setX(int iPart, MatrixXd tempX){swarm[iPart]->setX(tempX); }
    void setP(int iPart, VectorXd tempP){swarm[iPart]->setP(tempP); }
    void checkDuplicatePoints(int iPart){swarm[iPart]->checkDuplicatePoints();}
    void frameArrange(int iPart){swarm[iPart]->frameArrange();}
    
    
    // Printing Methods
    void displayWorstInformation();
    void printGbestDesign();
    
    // Saving methods
    void saveGbestDesign(string filepath, string prefix);
    void saveGbestLog(string filepath, string filename);
    void writePbestLogToCSV(string filepath, string filename);
    void writeFitnessLogToCSV(string filepath, string filename);
    
    ~qswarm();

};

#endif /* qswarm_hpp */
